import numpy as np
import pytest

from larray.tests.common import assert_array_equal, assert_axis_eq
from larray import Axis, AxisCollection


lipro = Axis('lipro=P01..P04')
sex = Axis('sex=M,F')
sex2 = Axis('sex=F,M')
age = Axis('age=0..7')
geo = Axis('geo=A11,A12,A13')
value = Axis('value=0..10')


@pytest.fixture
def col():
    return AxisCollection((lipro, sex, age))


def test_init_from_group():
    lipro_subset = lipro[:'P03']
    col2 = AxisCollection((lipro_subset, sex))
    assert col2.names == ['lipro', 'sex']
    assert_array_equal(col2.lipro.labels, ['P01', 'P02', 'P03'])
    assert_array_equal(col2.sex.labels, ['M', 'F'])


def test_init_from_string():
    col = AxisCollection('age=10;sex=M,F;year=2000..2017')
    assert col.names == ['age', 'sex', 'year']
    assert list(col.age.labels) == [10]
    assert list(col.sex.labels) == ['M', 'F']
    assert list(col.year.labels) == [y for y in range(2000, 2018)]


def test_eq(col):
    assert col == col
    assert col == AxisCollection((lipro, sex, age))
    assert col == (lipro, sex, age)
    assert col != (lipro, age, sex)


def test_getitem_name(col):
    assert_axis_eq(col['lipro'], lipro)
    assert_axis_eq(col['sex'], sex)
    assert_axis_eq(col['age'], age)


def test_getitem_int(col):
    assert_axis_eq(col[0], lipro)
    assert_axis_eq(col[-3], lipro)
    assert_axis_eq(col[1], sex)
    assert_axis_eq(col[-2], sex)
    assert_axis_eq(col[2], age)
    assert_axis_eq(col[-1], age)

    # test using a numpy int
    assert_axis_eq(col[np.int32(0)], lipro)


def test_getitem_slice(col):
    col = col[:2]
    assert len(col) == 2
    assert_axis_eq(col[0], lipro)
    assert_axis_eq(col[1], sex)


def test_setitem_name(col):
    col2 = col[:]
    col2['lipro'] = geo
    assert len(col2) == 3
    assert col2 == [geo, sex, age]
    col2['sex'] = sex2
    assert col2 == [geo, sex2, age]
    col2['geo'] = lipro
    assert col2 == [lipro, sex2, age]
    col2['age'] = geo
    assert col2 == [lipro, sex2, geo]
    col2['sex'] = sex
    col2['geo'] = age
    assert col2 == col


def test_setitem_name_axis_def(col):
    col2 = col[:]
    col2['lipro'] = 'geo=A11,A12,A13'
    assert len(col2) == 3
    assert col2 == [geo, sex, age]
    col2['sex'] = 'sex=F,M'
    assert col2 == [geo, sex2, age]
    col2['geo'] = 'lipro=P01..P04'
    assert col2 == [lipro, sex2, age]
    col2['age'] = 'geo=A11,A12,A13'
    assert col2 == [lipro, sex2, geo]
    col2['sex'] = 'sex=M,F'
    col2['geo'] = 'age=0..7'
    assert col2 == col


def test_setitem_int(col):
    col[1] = geo
    assert len(col) == 3
    assert col == [lipro, geo, age]
    col[2] = sex
    assert col == [lipro, geo, sex]
    col[-1] = age
    assert col == [lipro, geo, age]


def test_setitem_list_replace(col):
    col[['lipro', 'age']] = [geo, lipro]
    assert col == [geo, sex, lipro]


def test_setitem_slice_replace(col):
    col2 = col[:]
    col2[1:] = [geo, sex]
    assert col2 == [lipro, geo, sex]
    col2[1:] = col[1:]
    assert col2 == col


def test_setitem_slice_insert(col):
    col[1:1] = [geo]
    assert col == [lipro, geo, sex, age]


def test_setitem_slice_delete(col):
    col[1:2] = []
    assert col == [lipro, age]
    col[0:1] = []
    assert col == [age]


def test_delitem(col):
    assert len(col) == 3
    del col[0]
    assert len(col) == 2
    assert_axis_eq(col[0], sex)
    assert_axis_eq(col[1], age)
    del col['age']
    assert len(col) == 1
    assert_axis_eq(col[0], sex)
    del col[sex]
    assert len(col) == 0


def test_delitem_slice(col):
    assert len(col) == 3
    del col[0:2]
    assert len(col) == 1
    assert col == [age]
    del col[:]
    assert len(col) == 0


def test_pop(col):
    lipro, sex, age = col
    assert len(col) == 3
    assert col.pop() is age
    assert len(col) == 2
    assert col[0] is lipro
    assert col[1] is sex
    assert col.pop() is sex
    assert len(col) == 1
    assert col[0] is lipro
    assert col.pop() is lipro
    assert len(col) == 0


def test_replace(col):
    col2 = col[:]
    newcol = col2.replace('sex', geo)
    assert col2 == col
    assert len(newcol) == 3
    assert newcol.names == ['lipro', 'geo', 'age']
    assert newcol.shape == (4, 3, 8)
    newcol = newcol.replace(geo, sex)
    assert len(newcol) == 3
    assert newcol.names == ['lipro', 'sex', 'age']
    assert newcol.shape == (4, 2, 8)
    newcol = col2.replace(sex, 3)
    assert len(newcol) == 3
    assert newcol.names == ['lipro', None, 'age']
    assert newcol.shape == (4, 3, 8)
    newcol = col2.replace(sex, ['a', 'b', 'c'])
    assert len(newcol) == 3
    assert newcol.names == ['lipro', None, 'age']
    assert newcol.shape == (4, 3, 8)
    newcol = col2.replace(sex, "letters=a,b,c")
    assert len(newcol) == 3
    assert newcol.names == ['lipro', 'letters', 'age']
    assert newcol.shape == (4, 3, 8)


def test_contains(col):
    assert 'lipro' in col
    assert not ('nonexisting' in col)
    assert 0 in col
    assert 1 in col
    assert 2 in col
    assert -1 in col
    assert -2 in col
    assert -3 in col
    assert not (3 in col)
    assert lipro in col
    assert sex in col
    assert age in col
    assert sex2 in col
    assert not (geo in col)
    assert not (value in col)
    anon = Axis([0, 1])
    col.append(anon)
    assert anon in col
    anon2 = anon.copy()
    assert anon2 in col
    anon3 = Axis([0, 2])
    assert not (anon3 in col)


def test_index(col):
    assert col.index('lipro') == 0
    with pytest.raises(ValueError):
        col.index('nonexisting')
        assert col.index(0) == 0
    assert col.index(1) == 1
    assert col.index(2) == 2
    assert col.index(-1) == -1
    assert col.index(-2) == -2
    assert col.index(-3) == -3
    with pytest.raises(ValueError):
        col.index(3)

    # objects actually in col
    assert col.index(lipro) == 0
    assert col.index(sex) == 1
    assert col.index(age) == 2
    assert col.index(sex2) == 1
    with pytest.raises(ValueError):
        col.index(geo)
    with pytest.raises(ValueError):
        col.index(value)

    # test anonymous axes
    anon = Axis([0, 1])
    col.append(anon)
    assert col.index(anon) == 3
    anon2 = anon.copy()
    assert col.index(anon2) == 3
    anon3 = Axis([0, 2])
    with pytest.raises(ValueError):
        col.index(anon3)


def test_get(col):
    assert_axis_eq(col.get('lipro'), lipro)
    assert col.get('nonexisting') is None
    assert col.get('nonexisting', value) is value


def test_keys(col):
    assert col.keys() == ['lipro', 'sex', 'age']


def test_getattr(col):
    assert_axis_eq(col.lipro, lipro)
    assert_axis_eq(col.sex, sex)
    assert_axis_eq(col.age, age)


def test_append(col):
    geo = Axis('geo=A11,A12,A13')
    col.append(geo)
    assert col == [lipro, sex, age, geo]


def test_extend(col):
    col.extend([geo, value])
    assert col == [lipro, sex, age, geo, value]


def test_insert(col):
    col.insert(1, geo)
    assert col == [lipro, geo, sex, age]


def test_add(col):
    col2 = col.copy()
    new = col2 + [geo, value]
    assert new == [lipro, sex, age, geo, value]
    assert col2 == col
    new = col2 + [Axis('geo=A11,A12,A13'), Axis('age=0..7')]
    assert new == [lipro, sex, age, geo]
    with pytest.raises(ValueError):
        col2 + [Axis('geo=A11,A12,A13'), Axis('age=0..6')]

    # 2) other AxisCollection
    new = col2 + AxisCollection([geo, value])
    assert new == [lipro, sex, age, geo, value]


# def test_or():
#     col = AxisCollection('a0,a1')
#     res = col | Axis(2)
#     assert res == col
#
#     res = col | Axis(3)
#     assert res == AxisCollection([Axis('a0,a1'), Axis(3)])
#
#     res = col | AxisCollection([Axis(2), Axis(2)])
#     assert res == AxisCollection([Axis('a0,a1'), Axis(2)])


def test_sub():
    col = AxisCollection('a0,a1;b0,b1,b2')
    res = col - Axis(2)
    assert res == AxisCollection('b0,b1,b2')

    res = col - Axis(3)
    assert res == AxisCollection('a0,a1')

    col = AxisCollection('a0,a1;b0,b1')
    # when several axes are compatible, remove first
    res = col - Axis(2)
    assert res == AxisCollection('b0,b1')

    # when no axis is compatible, do not remove any
    res = col - Axis(3)
    assert res == col


def test_combine(col):
    res = col.combine_axes((lipro, sex))
    assert res.names == ['lipro_sex', 'age']
    assert res.size == col.size
    assert res.shape == (4 * 2, 8)
    print(res.info)
    assert_array_equal(res.lipro_sex.labels[0], 'P01_M')
    res = col.combine_axes((lipro, age))
    assert res.names == ['lipro_age', 'sex']
    assert res.size == col.size
    assert res.shape == (4 * 8, 2)
    assert_array_equal(res.lipro_age.labels[0], 'P01_0')
    res = col.combine_axes((sex, age))
    assert res.names == ['lipro', 'sex_age']
    assert res.size == col.size
    assert res.shape == (4, 2 * 8)
    assert_array_equal(res.sex_age.labels[0], 'M_0')


def test_info(col):
    expected = """\
4 x 2 x 8
 lipro [4]: 'P01' 'P02' 'P03' 'P04'
 sex [2]: 'M' 'F'
 age [8]: 0 1 2 ... 5 6 7"""
    assert col.info == expected


def test_str(col):
    assert str(col) == "{lipro, sex, age}"


def test_repr(col):
    assert repr(col) == """AxisCollection([
    Axis(['P01', 'P02', 'P03', 'P04'], 'lipro'),
    Axis(['M', 'F'], 'sex'),
    Axis([0, 1, 2, 3, 4, 5, 6, 7], 'age')
])"""


if __name__ == "__main__":
    pytest.main()



# numpy ufuncs
# http://docs.scipy.org/doc/numpy/reference/routines.math.html
import numpy as np

from larray.core.array import Array, make_args_broadcastable


def wrap_elementwise_array_func(func, doc=None):
    r"""
    Wrap a function using numpy arrays to work with LArray arrays instead.

    Parameters
    ----------
    func : function
        A function taking numpy arrays as arguments and returning numpy arrays of the same shape. If the function
        takes several arguments, this wrapping code assumes the result will have the combination of all axes present.
        In numpy talk, arguments will be broadcasted to each other.
    doc : str, optional
        The documentation (docstring) for the new function. Defaults to the documentation of the original function,
        if any.

    Returns
    -------
    function
        A function taking larray.Array arguments and returning larray.Arrays.

    Examples
    --------
    For example, if we want to apply the Hodrick-Prescott filter from statsmodels we can use this:

    >>> from statsmodels.tsa.filters.hp_filter import hpfilter         # doctest: +SKIP
    >>> hpfilter = wrap_elementwise_array_func(hpfilter)               # doctest: +SKIP

    hpfilter is now a function taking a one dimensional Array as input and returning a one dimensional Array as output

    Now let us suppose we have a ND array such as:

    >>> from larray.random import normal
    >>> arr = normal(axes="sex=M,F;year=2016..2018")                   # doctest: +SKIP
    >>> arr                                                            # doctest: +SKIP
    sex\year   2016   2017   2018
           M  -1.15   0.56  -1.06
           F  -0.48  -0.39  -0.98

    We can apply an Hodrick-Prescott filter to it by using:

    >>> # 6.25 is the recommended smoothing value for annual data
    >>> cycle, trend = arr.apply(hpfilter, 6.25, axes="year")          # doctest: +SKIP
    >>> trend                                                          # doctest: +SKIP
    sex\year   2016   2017   2018
           M  -0.61  -0.52  -0.52
           F  -0.37  -0.61  -0.87
    """
    def wrapper(*args, **kwargs):
        raw_bcast_args, raw_bcast_kwargs, res_axes = make_args_broadcastable(args, kwargs)

        # We pass only raw numpy arrays to the ufuncs even though numpy is normally meant to handle those cases itself
        # via __array_wrap__

        # There is a problem with np.clip though (and possibly other ufuncs): np.clip is roughly equivalent to
        # np.maximum(np.minimum(np.asarray(la), high), low)
        # the np.asarray(la) is problematic because it lose original labels
        # and then tries to get them back from high, where they are possibly
        # incomplete if broadcasting happened

        # It fails on "np.minimum(ndarray, Array)" because it calls __array_wrap__(high, result) which cannot work if
        # there was broadcasting involved (high has potentially less labels than result).
        # it does this because numpy calls __array_wrap__ on the argument with the highest __array_priority__
        res_data = func(*raw_bcast_args, **raw_bcast_kwargs)
        if res_axes:
            if isinstance(res_data, tuple):
                return tuple(Array(res_arr, res_axes) for res_arr in res_data)
            else:
                return Array(res_data, res_axes)
        else:
            return res_data
    # copy function name. We are intentionally not using functools.wraps, because it does not work for wrapping a
    # function from another module
    wrapper.__name__ = func.__name__
    wrapper.__doc__ = func.__doc__ if doc is None else doc
    return wrapper


def wrap_numpy_func(func, doc=None):
    # update documentation by inserting a warning message after the short description of the numpy function
    # (otherwise the description of ufuncs given in the corresponding API 'autosummary' tables will always
    #  start with 'larray specific variant of ...' without giving a meaningful description of what does the ufunc)
    if doc is None:
        if func.__doc__.startswith('\n'):
            # docstring starts with short description
            end_signature = 1
            end_short_desc = func.__doc__.find('\n\n')
        else:
            # docstring starts with signature
            end_signature = func.__doc__.find('\n\n') + 2
            end_short_desc = func.__doc__.find('\n\n', end_signature)
        short_desc = func.__doc__[:end_short_desc]
        numpy_doc = func.__doc__[end_short_desc:]
        ident = ' ' * (len(short_desc[end_signature:]) - len(short_desc[end_signature:].lstrip()))
        doc = f'{short_desc}\n\n{ident}larray specific variant of ``numpy.{func.__name__}``.\n\n' \
              f'{ident}Documentation from numpy:{numpy_doc}'
    wrapper = wrap_elementwise_array_func(func, doc)
    # set __qualname__ explicitly (all these functions are supposed to be top-level function in the ufuncs module)
    wrapper.__qualname__ = func.__name__
    # we should not copy __module__
    return wrapper


where = wrap_numpy_func(np.where, r"""
where(condition, x, y)

    Return elements, either from `x` or `y`, depending on `condition`.

    Parameters
    ----------
    condition : boolean Array
        When True, yield `x`, otherwise yield `y`.
    x, y : Array
        Values from which to choose.

    Returns
    -------
    out : Array
        If both `x` and `y` are specified, the output array contains
        elements of `x` where `condition` is True, and elements from
        `y` elsewhere.

    Examples
    --------
    >>> from larray import Array
    >>> arr = Array([[10, 7, 5, 9],
    ...               [5, 8, 3, 7],
    ...               [6, 2, 0, 9],
    ...               [9, 10, 5, 6]], "a=a0..a3;b=b0..b3")
    >>> arr
    a\b  b0  b1  b2  b3
     a0  10   7   5   9
     a1   5   8   3   7
     a2   6   2   0   9
     a3   9  10   5   6

    Simple use

    >>> where(arr <= 5, 0, arr)
    a\b  b0  b1  b2  b3
     a0  10   7   0   9
     a1   0   8   0   7
     a2   6   0   0   9
     a3   9  10   0   6

    With broadcasting

    >>> mean_by_col = arr.mean('a')
    >>> mean_by_col
    b   b0    b1    b2    b3
       7.5  6.75  3.25  7.75
    >>> # for each column, set values below the mean value to the mean value
    >>> where(arr < mean_by_col, mean_by_col, arr)
    a\b    b0    b1    b2    b3
     a0  10.0   7.0   5.0   9.0
     a1   7.5   8.0  3.25  7.75
     a2   7.5  6.75  3.25   9.0
     a3   9.0  10.0   5.0  7.75
""")

maximum = wrap_numpy_func(np.maximum, r"""
maximum(x1, x2, out=None, dtype=None)

    Element-wise maximum of array elements.

    Compare two arrays and returns a new array containing the element-wise
    maxima. If one of the elements being compared is a NaN, then that
    element is returned. If both elements are NaNs then the first is
    returned. The latter distinction is important for complex NaNs, which
    are defined as at least one of the real or imaginary parts being a NaN.
    The net effect is that NaNs are propagated.

    Parameters
    ----------
    x1, x2 : Array
        The arrays holding the elements to be compared.
    out : Array, optional
        An array into which the result is stored.
    dtype : data-type, optional
        Overrides the dtype of the output array.

    Returns
    -------
    y : Array or scalar
        The maximum of `x1` and `x2`, element-wise.
        This is a scalar if both `x1` and `x2` are scalars.

    See Also
    --------
    minimum :
        Element-wise minimum of two arrays, propagates NaNs.

    Notes
    -----
    The maximum is equivalent to ``where(x1 >= x2, x1, x2)`` when
    neither x1 nor x2 are NaNs, but it is faster.

    Examples
    --------
    >>> from larray import Array
    >>> arr1 = Array([[10, 7, 5, 9],
    ...                [5, 8, 3, 7]], "a=a0,a1;b=b0..b3")
    >>> arr2 = Array([[6, 2, 9, 0],
    ...                [9, 10, 5, 6]], "a=a0,a1;b=b0..b3")
    >>> arr1
    a\b  b0  b1  b2  b3
     a0  10   7   5   9
     a1   5   8   3   7
    >>> arr2
    a\b  b0  b1  b2  b3
     a0   6   2   9   0
     a1   9  10   5   6

    >>> maximum(arr1, arr2)
    a\b  b0  b1  b2  b3
     a0  10   7   9   9
     a1   9  10   5   7

    With broadcasting

    >>> arr2['a0']
    b  b0  b1  b2  b3
        6   2   9   0
    >>> maximum(arr1, arr2['a0'])
    a\b  b0  b1  b2  b3
     a0  10   7   9   9
     a1   6   8   9   7
""")

minimum = wrap_numpy_func(np.minimum, r"""
minimum(x1, x2, out=None, dtype=None)

    Element-wise minimum of array elements.

    Compare two arrays and returns a new array containing the element-wise
    minima. If one of the elements being compared is a NaN, then that
    element is returned. If both elements are NaNs then the first is
    returned. The latter distinction is important for complex NaNs, which
    are defined as at least one of the real or imaginary parts being a NaN.
    The net effect is that NaNs are propagated.

    Parameters
    ----------
    x1, x2 : Array
        The arrays holding the elements to be compared.
    out : Array, optional
        An array into which the result is stored.
    dtype : data-type, optional
        Overrides the dtype of the output array.

    Returns
    -------
    y : Array or scalar
        The minimum of `x1` and `x2`, element-wise.
        This is a scalar if both `x1` and `x2` are scalars.

    See Also
    --------
    maximum :
        Element-wise maximum of two arrays, propagates NaNs.

    Notes
    -----
    The minimum is equivalent to ``where(x1 <= x2, x1, x2)`` when
    neither x1 nor x2 are NaNs, but it is faster.

    Examples
    --------
    >>> from larray import Array
    >>> arr1 = Array([[10, 7, 5, 9],
    ...                [5, 8, 3, 7]], "a=a0,a1;b=b0..b3")
    >>> arr2 = Array([[6, 2, 9, 0],
    ...                [9, 10, 5, 6]], "a=a0,a1;b=b0..b3")
    >>> arr1
    a\b  b0  b1  b2  b3
     a0  10   7   5   9
     a1   5   8   3   7
    >>> arr2
    a\b  b0  b1  b2  b3
     a0   6   2   9   0
     a1   9  10   5   6

    >>> minimum(arr1, arr2)
    a\b  b0  b1  b2  b3
     a0   6   2   5   0
     a1   5   8   3   6

    With broadcasting

    >>> arr2['a0']
    b  b0  b1  b2  b3
        6   2   9   0
    >>> minimum(arr1, arr2['a0'])
    a\b  b0  b1  b2  b3
     a0   6   2   5   0
     a1   5   2   3   0
""")



from io import StringIO

from pandas import DataFrame, Index

from larray.core.constants import nan
from larray.util.misc import deprecate_kwarg
from larray.inout.common import _get_index_col
from larray.inout.pandas import df_asarray, set_dataframe_index_by_position
from larray.inout.csv import read_csv


@deprecate_kwarg('nb_index', 'nb_axes', arg_converter=lambda x: x + 1)
def from_lists(data, nb_axes=None, index_col=None, fill_value=nan, sort_rows=False, sort_columns=False, wide=True):
    r"""
    initialize array from a list of lists (lines)

    Parameters
    ----------
    data : sequence (tuple, list, ...)
        Input data. All data is supposed to already have the correct type (e.g. strings are not parsed).
    nb_axes : int, optional
        Number of axes of output array. The first `nb_axes` - 1 columns and the header will be used
        to set the axes of the output array. If not specified, the number of axes is given by the position of the
        column header including the character `\` plus one. If no column header includes the character `\`, the array
        is assumed to have one axis. Defaults to None.
    index_col : list, optional
        Positions of columns for the n-1 first axes (ex. [0, 1, 2, 3]). Defaults to None (see nb_axes above).
    fill_value : scalar or Array, optional
        Value used to fill cells corresponding to label combinations which are not present in the input.
        Defaults to NaN.
    sort_rows : bool, optional
        Whether or not to sort the rows alphabetically (sorting is more efficient than not sorting). Defaults to False.
    sort_columns : bool, optional
        Whether or not to sort the columns alphabetically (sorting is more efficient than not sorting).
        Defaults to False.
    wide : bool, optional
        Whether or not to assume the array is stored in "wide" format.
        If False, the array is assumed to be stored in "narrow" format: one column per axis plus one value column.
        Defaults to True.

    Returns
    -------
    Array

    Examples
    --------
    >>> from_lists([['sex', 'M', 'F'],
    ...             ['',      0,   1]])
    sex  M  F
         0  1
    >>> from_lists([['sex\\year', 1991, 1992, 1993],
    ...             [ 'M',           0,    1,    2],
    ...             [ 'F',           3,    4,    5]])
    sex\year  1991  1992  1993
           M     0     1     2
           F     3     4     5

    Read array with missing values + `fill_value` argument

    >>> from_lists([['sex', 'nat\\year', 1991, 1992, 1993],
    ...             [  'M', 'BE',           1,    0,    0],
    ...             [  'M', 'FO',           2,    0,    0],
    ...             [  'F', 'BE',           0,    0,    1]])
    sex  nat\year  1991  1992  1993
      M        BE   1.0   0.0   0.0
      M        FO   2.0   0.0   0.0
      F        BE   0.0   0.0   1.0
      F        FO   nan   nan   nan

    >>> from_lists([['sex', 'nat\\year', 1991, 1992, 1993],
    ...             [  'M', 'BE',           1,    0,    0],
    ...             [  'M', 'FO',           2,    0,    0],
    ...             [  'F', 'BE',           0,    0,    1]], fill_value=42)
    sex  nat\year  1991  1992  1993
      M        BE     1     0     0
      M        FO     2     0     0
      F        BE     0     0     1
      F        FO    42    42    42

    Specify the number of axes of the array to be read

    >>> from_lists([['sex', 'nat', 1991, 1992, 1993],
    ...             [  'M', 'BE',     1,    0,    0],
    ...             [  'M', 'FO',     2,    0,    0],
    ...             [  'F', 'BE',     0,    0,    1]], nb_axes=3)
    sex  nat\{2}  1991  1992  1993
      M       BE   1.0   0.0   0.0
      M       FO   2.0   0.0   0.0
      F       BE   0.0   0.0   1.0
      F       FO   nan   nan   nan

    Read array saved in "narrow" format (wide=False)

    >>> from_lists([['sex', 'nat', 'year', 'value'],
    ...             [  'M', 'BE',  1991,    1     ],
    ...             [  'M', 'BE',  1992,    0     ],
    ...             [  'M', 'BE',  1993,    0     ],
    ...             [  'M', 'FO',  1991,    2     ],
    ...             [  'M', 'FO',  1992,    0     ],
    ...             [  'M', 'FO',  1993,    0     ],
    ...             [  'F', 'BE',  1991,    0     ],
    ...             [  'F', 'BE',  1992,    0     ],
    ...             [  'F', 'BE',  1993,    1     ]], wide=False)
    sex  nat\year  1991  1992  1993
      M        BE   1.0   0.0   0.0
      M        FO   2.0   0.0   0.0
      F        BE   0.0   0.0   1.0
      F        FO   nan   nan   nan
    """
    index_col = _get_index_col(nb_axes, index_col, wide)

    columns = data[0]
    # issue #950: avoid pandas interpreting [None, 0, 1] as a Float64Index([nan, 0.0, 1.0], dtype='float64')
    if None in columns:
        columns = Index(columns, dtype=object)

    df = DataFrame(data[1:], columns=columns)
    if index_col is not None:
        df = set_dataframe_index_by_position(df, index_col)

    return df_asarray(df, raw=index_col is None, parse_header=False, sort_rows=sort_rows, sort_columns=sort_columns,
                      fill_value=fill_value, wide=wide)


@deprecate_kwarg('nb_index', 'nb_axes', arg_converter=lambda x: x + 1)
def from_string(s, nb_axes=None, index_col=None, sep=' ', wide=True, **kwargs):
    r"""Create an array from a multi-line string.

    Parameters
    ----------
    s : str
        input string.
    nb_axes : int, optional
        Number of axes of output array. The first `nb_axes` - 1 columns and the header will be used
        to set the axes of the output array. If not specified, the number of axes is given by the position of the
        column header including the character `\` plus one. If no column header includes the character `\`, the array
        is assumed to have one axis. Defaults to None.
    index_col : list, optional
        Positions of columns for the n-1 first axes (ex. [0, 1, 2, 3]). Defaults to None (see nb_axes above).
    sep : str
        delimiter used to split each line into cells.
    wide : bool, optional
        Whether or not to assume the array is stored in "wide" format.
        If False, the array is assumed to be stored in "narrow" format: one column per axis plus one value column.
        Defaults to True.
    **kwargs
        See arguments of Pandas read_csv function.

    Returns
    -------
    Array

    Examples
    --------
    >>> # to create a 1D array using the default separator ' ', a tabulation character \t must be added in front
    >>> # of the data line
    >>> from_string("sex  M  F\n\t  0  1")
    sex  M  F
         0  1
    >>> from_string("nat\\sex  M  F\nBE  0  1\nFO  2  3")
    nat\sex  M  F
         BE  0  1
         FO  2  3
    >>> from_string("period  a  b\n2010  0  1\n2011  2  3")
    period\{1}  a  b
          2010  0  1
          2011  2  3

    Each label is stripped of leading and trailing whitespace, so this is valid too:

    >>> from_string('''nat\\sex  M  F
    ...                BE        0  1
    ...                FO        2  3''')
    nat\sex  M  F
         BE  0  1
         FO  2  3
    >>> from_string('''age  nat\\sex  M  F
    ...                0    BE        0  1
    ...                0    FO        2  3
    ...                1    BE        4  5
    ...                1    FO        6  7''')
    age  nat\sex  M  F
      0       BE  0  1
      0       FO  2  3
      1       BE  4  5
      1       FO  6  7

    Empty lines at the beginning or end are ignored, so one can also format the string like this:

    >>> from_string('''
    ... nat\\sex  M  F
    ... BE        0  1
    ... FO        2  3
    ... ''')
    nat\sex  M  F
         BE  0  1
         FO  2  3
    """
    return read_csv(StringIO(s), nb_axes=nb_axes, index_col=index_col, sep=sep, skipinitialspace=True,
                    wide=wide, **kwargs)



# define abstract base classes to enable isinstance type checking on our objects
# idea taken from https://github.com/pandas-dev/pandas/blob/master/pandas/core/dtypes/generic.py
# we do not inherit from abc.ABC because it costs us a ~3% performance bump on
# our benchmarks for very little benefit
class ABCAxis:
    pass


class ABCAxisReference(ABCAxis):
    pass


class ABCArray:
    pass



import os
from datetime import date, time, datetime
from collections import OrderedDict

from larray.core.axis import Axis
from larray.core.group import Group
from larray.core.array import Array


# all formats
_supported_larray_types = (Axis, Group, Array)

# only for HDF5 and pickle formats
# support list, tuple and dict?
_supported_scalars_types = (int, float, bool, bytes, str, date, time, datetime)
_supported_types = _supported_larray_types + _supported_scalars_types
_supported_typenames = {cls.__name__ for cls in _supported_types}


def _get_index_col(nb_axes=None, index_col=None, wide=True):
    if not wide:
        if nb_axes is not None or index_col is not None:
            raise ValueError("`nb_axes` or `index_col` argument cannot be used when `wide` argument is False")

    if nb_axes is not None and index_col is not None:
        raise ValueError("cannot specify both `nb_axes` and `index_col`")
    elif nb_axes is not None:
        index_col = list(range(nb_axes - 1))
    elif isinstance(index_col, int):
        index_col = [index_col]

    return index_col


class FileHandler:
    r"""
    Abstract class defining the methods for "file handler" subclasses.

    Parameters
    ----------
    fname : str
        Filename.

    Attributes
    ----------
    fname : str
        Filename.
    """
    def __init__(self, fname, overwrite_file=False):
        self.fname = fname
        self.original_file_name = None
        self.overwrite_file = overwrite_file

    def _open_for_read(self):
        raise NotImplementedError()

    def _open_for_write(self):
        raise NotImplementedError()

    def list_items(self):
        r"""
        Return list containing pairs (name, type) for all stored objects
        """
        raise NotImplementedError()

    def _read_item(self, key, type, *args, **kwargs):
        r"""Read item"""
        raise NotImplementedError()

    def _read_metadata(self):
        r"""Read metadata"""
        raise NotImplementedError()

    def _dump_item(self, key, value, *args, **kwargs):
        r"""Dump item. Raises an TypeError if type not taken into account by the FileHandler subclass."""
        raise NotImplementedError()

    def _dump_metadata(self, metadata):
        r"""Dump metadata"""
        raise NotImplementedError()

    def save(self):
        r"""
        Saves items in file.
        """
        pass

    def close(self):
        r"""
        Closes file.
        """
        raise NotImplementedError()

    def _get_original_file_name(self):
        if self.overwrite_file and os.path.isfile(self.fname):
            self.original_file_name = self.fname
            fname, ext = os.path.splitext(self.fname)
            self.fname = f'{fname}~{ext}'

    def _update_original_file(self):
        if self.original_file_name is not None and os.path.isfile(self.fname):
            os.remove(self.original_file_name)
            os.rename(self.fname, self.original_file_name)

    def read(self, keys, *args, **kwargs):
        r"""
        Reads file content (HDF, Excel, CSV, ...) and returns a dictionary containing loaded objects.

        Parameters
        ----------
        keys : list of str
            List of objects' names.
        *args : any
            Any other argument is passed through to the underlying read function.
        display : bool, optional
            Whether or not the function should display a message when starting and ending to load each object.
            Defaults to False.
        ignore_exceptions : bool, optional
            Whether or not an exception should stop the function or be ignored. Defaults to False.
        **kwargs : any
            Any other keyword argument is passed through to the underlying read function.

        Returns
        -------
        Metadata
            List of metadata to load.
        OrderedDict(str, Array/Axis/Group)
            Dictionary containing the loaded objects.
        """
        display = kwargs.pop('display', False)
        ignore_exceptions = kwargs.pop('ignore_exceptions', False)
        self._open_for_read()
        metadata = self._read_metadata()
        key_types = self.list_items()
        if keys is not None:
            key_types = [(key, type) for key, type in key_types if key in keys]
        res = OrderedDict()
        for key, type in key_types:
            if display:
                print("loading", type, "object", key, "...", end=' ')
            try:
                res[key] = self._read_item(key, type, *args, **kwargs)
            except Exception:
                if not ignore_exceptions:
                    raise
            if display:
                print("done")
        self.close()
        return metadata, res

    def dump(self, metadata, key_values, *args, **kwargs):
        r"""
        Dumps objects corresponding to keys in file in HDF, Excel, CSV, ... format

        Parameters
        ----------
        metadata: Metadata
            List of metadata to dump.
        key_values : list of (str, Array/Axis/Group) pairs
            Name and data of objects to dump.
        kwargs :
            * display: whether or not to display when the dump of each object is started/done.
        """
        display = kwargs.pop('display', False)
        self._get_original_file_name()
        self._open_for_write()
        if metadata is not None:
            self._dump_metadata(metadata)
        for key, value in key_values:
            if isinstance(value, Array) and value.ndim == 0:
                if display:
                    print(f'Cannot dump {key}. Dumping 0D arrays is currently not supported.')
                continue
            try:
                if display:
                    print("dumping", key, "...", end=' ')
                self._dump_item(key, value, *args, **kwargs)
                if display:
                    print("done")
            except TypeError:
                if display:
                    print(f"Cannot dump {key}. {type(value).__name__} is not a supported type")
        self.save()
        self.close()
        self._update_original_file()



import math
from collections import deque

from larray.core.array import Array, asarray, ones, any
import numpy as np


def badvalues(a, bad_filter):
    bad_values = a[bad_filter]
    assert bad_values.ndim == 1
    return '\n'.join(f'{k}: {v}' for k, v in zip(bad_values.axes[0], bad_values))


def f2str(f, threshold=2):
    r"""Return string representation of floating point number f.
    Use scientific notation if f would have more than threshold decimal digits, otherwise use threshold as precision.

    Parameters
    ----------
    f : float
        Number to represent.
    threshold : int, optional
        Precision (number of decimal digits displayed). If the number needs more digits, scientific notation will be
        used.

    Examples
    --------
    >>> f2str(55.1)
    '55.10'
    >>> f2str(1.234)
    '1.23'
    >>> f2str(0.002)
    '2.00e-03'
    """
    kind = "e" if f and math.log10(1 / abs(f)) > threshold else "f"
    return f"{f:.{threshold}{kind}}"


def warn_or_raise(what, msg):
    if what == 'raise':
        raise ValueError(msg)
    else:
        print(f"WARNING: {msg}")


def ipfp(target_sums, a=None, axes=None, maxiter=1000, threshold=0.5, stepstoabort=10, nzvzs='raise',
         no_convergence='raise', display_progress=False):
    r"""Apply Iterative Proportional Fitting Procedure (also known as bi-proportional fitting in statistics,
    RAS algorithm in economics) to array a, with target_sums as targets.

    Parameters
    ----------
    target_sums : tuple/list of array-like
        Target sums to achieve.
        First element must be the sum to achieve along axis 0, the second the sum along axis 1, ...
    a : array-like, optional
        Starting values to fit, if not given starts with an array filled with 1.
    axes : list/tuple of axes, optional
        Axes on which the fitting procedure should be applied. Defaults to all axes.
    maxiter : int, optional
        Maximum number of iteration, defaults to 1000.
    threshold : float, optional
        Threshold below which the result is deemed acceptable, defaults to 0.5.
    stepstoabort : int, optional
        Number of consecutive steps with no improvement after which to abort. Defaults to 10.
    nzvzs : 'fix', 'warn' or 'raise', optional
        Behavior when detecting non zero values where the sum is zero
        'fix': set to zero (silently)
        'warn': set to zero and print a warning
        'raise': raise an exception (default)
    no_convergence : 'ignore', 'warn' or 'raise, optional
        Behavior when the algorithm does not seem to converge. This condition is triggered both when the maximum number
        of iteration is reached or when the maximum absolute difference between the target and the current sums does
        not improve for `stepstoabort` iterations.
        'ignore': return values computed up to that point (silently)
        'warn': return values computed up to that point and print a warning
        'raise': raise an exception (default)
    display_progress : False, True or 'condensed', optional
        Whether or not to display progress. Defaults to False.
        If 'condensed' will display progress using a denser template (using one line per iteration).

    Returns
    -------
    Array

    Examples
    --------
    >>> from larray import *
    >>> a = Axis('a=a0,a1')
    >>> b = Axis('b=b0,b1')
    >>> initial = Array([[2, 1], [1, 2]], [a, b])
    >>> initial
    a\b  b0  b1
     a0   2   1
     a1   1   2
    >>> target_sum_along_a = Array([2, 1], b)
    >>> target_sum_along_a
    b  b0  b1
        2   1
    >>> target_sum_along_b = Array([1, 2], a)
    >>> target_sum_along_b
    a  a0  a1
        1   2
    >>> result = ipfp([target_sum_along_a, target_sum_along_b], initial, threshold=0.01)
    >>> # round result so that its display is nicer
    ... round(result, 2)
    a\b    b0    b1
     a0  0.85  0.15
     a1  1.15  0.85

    Now let us assume you have a 3D array like this:

    >>> year = Axis('year=2014..2016')
    >>> initial = ndtest([a, b, year])
    >>> initial
     a  b\year  2014  2015  2016
    a0      b0     0     1     2
    a0      b1     3     4     5
    a1      b0     6     7     8
    a1      b1     9    10    11

    and some targets for each year:

    >>> btargets = initial.sum(X.a) + 1
    >>> btargets
    b\year  2014  2015  2016
        b0     7     9    11
        b1    13    15    17
    >>> atargets = initial.sum(X.b) + 1
    >>> atargets
    a\year  2014  2015  2016
        a0     4     6     8
        a1    16    18    20

    You want to apply a 2D fitting procedure for each value of that year axis. You could call ipfp within a loop on
    the year axis, but you can also apply the procedure for all years at once by using the axes argument. This is
    *much* faster than an explicit loop.

    >>> result = ipfp([btargets, atargets], initial, axes=(X.a, X.b))
    """
    assert nzvzs in {'fix', 'warn', 'raise'}
    assert no_convergence in {'ignore', 'warn', 'raise'}
    assert isinstance(display_progress, bool) or display_progress == 'condensed'

    target_sums = [asarray(ts) for ts in target_sums]

    n = len(target_sums)

    if axes is None:
        axes = list(range(n))

    def has_anonymous_axes(a):
        return any(axis.name is None for axis in a.axes)

    if any(has_anonymous_axes(ts) for ts in target_sums):
        if any(not isinstance(axis, int) for axis in axes):
            raise ValueError("ipfp does not support target sums with anonymous axes when using the axes argument with"
                             "non-integer (positional) axis references")

        names_for_missing_axes = [f'axis{i}' for i in axes]
        new_target_sums = []
        for i, target_sum in zip(axes, target_sums):
            ts_axes_names = names_for_missing_axes[:i] + names_for_missing_axes[i + 1:]
            new_ts = target_sum.rename({axis: name
                                        for axis, name in zip(target_sum.axes, ts_axes_names)
                                        if axis.name is None})
            new_target_sums.append(new_ts)
        target_sums = new_target_sums

    if a is None:
        # reconstruct all axes from target_sums axes
        # assuming shape is the total shape of a
        # >>> shape = (3, 4, 5)
        # target_sums axes would be:
        # >>> shapes = [shape[:i] + shape[i+1:] for i in range(len(shape))]
        # >>> shapes
        # [(4, 5), (3, 5), (3, 4)]
        # >>> (shapes[1][0],) + shapes[0]
        # (3, 4, 5)
        # so, to reconstruct a.axes from target_sum axes, we need to take the first axis of the second target_sum and
        # all axes from the first target_sum:
        first_axis = target_sums[1].axes[0]
        other_axes = target_sums[0].axes
        all_axes = first_axis + other_axes
        a = ones(all_axes, dtype=np.float64)
    else:
        # TODO: only make a copy if there are actually any bad values, but I am unsure we should make a copy at all.
        # Either way, this should be documented.
        if nzvzs in {'warn', 'fix'} and isinstance(a, Array):
            a = a.copy()
        else:
            a = asarray(a)
        # TODO: this should be a builtin op
        a = a.rename({i: name if name is not None else f'axis{i}'
                      for i, name in enumerate(a.axes.names)})

    axes = a.axes[axes]

    # this test should only ever fail if the user passed larray for a and target sums
    for axis, axis_target_sum in zip(axes, target_sums):
        expected_axes = a.axes - axis
        if axis_target_sum.axes != expected_axes:
            raise ValueError(f"axes of target sum along {axis.name} (axis {a.axes.index(axis)}) do not match "
                             f"corresponding array axes: got {axis_target_sum.axes} but expected {expected_axes}. "
                             f"Are the target sums in the correct order?")

    axis0_total = target_sums[0].sum()
    for axis, axis_target_sum in zip(axes[1:], target_sums[1:]):
        axis_total = axis_target_sum.sum()
        if str(axis_total) != str(axis0_total):
            raise ValueError(f"target sum along {axis} (axis {a.axes.index(axis)}) is different than target sum along "
                             f"{axes[0]} (axis {a.axes.index(axes[0])}): {axis_total} vs {axis0_total}")

    negative = a < 0
    if any(negative):
        raise ValueError(f"negative value(s) found:\n{badvalues(a, negative)}")

    for axis, axis_target_sum in zip(axes, target_sums):
        axis_idx = a.axes.index(axis)
        axis_sum = a.sum(axis)
        bad = (axis_sum == 0) & (axis_target_sum != 0)
        if any(bad):
            raise ValueError(f"found all zero values sum along {axis.name} (axis {axis_idx}) but non zero target sum:\n"
                             f"{badvalues(axis_target_sum, bad)}")

        bad = (axis_sum != 0) & (axis_target_sum == 0)
        if any(bad):
            if nzvzs in {'warn', 'raise'}:
                msg = f"found Non Zero Values but Zero target Sum (nzvzs) along {axis.name} (axis {axis_idx})"
                if nzvzs == 'raise':
                    raise ValueError(f"{msg}, use nzvzs='warn' or 'fix' to set them to zero automatically:\n"
                                     f"{badvalues(axis_sum, bad)}")
                else:
                    print(f"WARNING: {msg}, setting them to zero:\n{badvalues(axis_sum, bad)}")

            a[bad] = 0
            # verify we did fix the problem
            assert not any((a.sum(axis) != 0) & (axis_target_sum == 0))

    r = a
    lastdiffs = deque([float('nan')], maxlen=stepstoabort)

    # Here is the nice version of the algorithm

    # for i in range(maxiter):
    #     startr = r
    #     for axis, axis_target in zip(axes, target_sums):
    #         r = r * axis_target.divnot0(r.sum(axis))
    #     max_sum_diff = max(abs(r.sum(axis) - axis_target).max()
    #                        for axis, axis_target in zip(axes, target_sums))
    #     step_sum_improvement = ...

    # Here is the ugly optimized version which avoids computing the sum for the first axis twice per iteration
    # (saves ~10-15% running time).
    axis0_sum = r.sum(axes[0])
    for i in range(maxiter):
        startr = r
        r = r * target_sums[0].divnot0(axis0_sum)
        for axis, axis_target in zip(axes[1:], target_sums[1:]):
            r = r * axis_target.divnot0(r.sum(axis))

        axes_sum = [r.sum(axis) for axis in axes]
        max_sum_diff = max(abs(axis_sum - axis_target).max()
                           for axis_sum, axis_target in zip(axes_sum, target_sums))
        axis0_sum = axes_sum[0]

        step_sum_improvement = lastdiffs[-1] - max_sum_diff
        stepcelldiff = abs(r - startr).max()

        if display_progress:
            maxcelldiff = f2str(stepcelldiff)
            maxdiff2target = f2str(max_sum_diff)
            stepchange = f2str(step_sum_improvement)
            if display_progress == "condensed":
                print(f"it {i} max cell diff {maxcelldiff} max diff to target {maxdiff2target} "
                      f"({stepchange})")
            else:
                print(f"""iteration {i}
 * max(abs(prev_cell - cell)): {maxcelldiff}
 * max(abs(sum - target_sum)): {maxdiff2target}
   \\- change since last iteration: {stepchange}
""")

        if np.all(np.array(lastdiffs) == max_sum_diff):
            if no_convergence in {'warn', 'raise'}:
                warn_or_raise(no_convergence, f"does not seem to converge (no improvement for {stepstoabort} "
                                              f"consecutive steps), stopping here.")
            return r

        if max_sum_diff < threshold:
            if display_progress:
                print(f"acceptable max(abs(sum - target_sum)) found at iteration {i}: "
                      f"{f2str(max_sum_diff)} < threshold ({threshold})")
            return r

        lastdiffs.append(max_sum_diff)

    if no_convergence in {'warn', 'raise'}:
        warn_or_raise(no_convergence, f"maximum iteration reached ({maxiter})")
    return r



import warnings

import numpy as np
import pandas as pd

from larray.core.constants import nan
from larray.inout.pandas import df_asarray
from larray.util.misc import deprecate_kwarg


@deprecate_kwarg('nb_index', 'nb_axes', arg_converter=lambda x: x + 1)
def read_sas(filepath, nb_axes=None, index_col=None, fill_value=nan, na=nan, sort_rows=False, sort_columns=False,
             **kwargs):
    r"""
    Reads sas file and returns an Array with the contents
        nb_axes: number of axes of the output array
    or
        index_col: Positions of columns for the n-1 first axes (ex. [0, 1, 2, 3])
    """
    if not np.isnan(na):
        fill_value = na
        warnings.warn("read_sas `na` argument has been renamed to `fill_value`. Please use that instead.",
                      FutureWarning, stacklevel=2)

    if nb_axes is not None and index_col is not None:
        raise ValueError("cannot specify both nb_axes and index_col")
    elif nb_axes is not None:
        index_col = list(range(nb_axes - 1))
    elif isinstance(index_col, int):
        index_col = [index_col]

    df = pd.read_sas(filepath, index=index_col, **kwargs)
    return df_asarray(df, sort_rows=sort_rows, sort_columns=sort_columns, fill_value=fill_value)



import os
from setuptools import setup, find_packages


def readlocal(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


DISTNAME = 'larray'
VERSION = '0.33'
AUTHOR = 'Gaetan de Menten, Geert Bryon, Johan Duyck, Alix Damman'
AUTHOR_EMAIL = 'gdementen@gmail.com'
DESCRIPTION = "N-D labeled arrays in Python"
LONG_DESCRIPTION = readlocal("README.rst")
SETUP_REQUIRES = []
# - pandas >= 0.20.0 is required since commit 01669f2024a7bffe47cceec0a0fd845f71b6f7cc
#   (issue 702 : fixed bug when writing metadata using HDF format)
INSTALL_REQUIRES = ['numpy >= 1.13', 'pandas >= 0.20.0']
TESTS_REQUIRE = ['pytest', 'pytest-flake8']

LICENSE = 'GPLv3'
URL = 'https://github.com/larray-project/larray'

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    'Operating System :: OS Independent',
    'Intended Audience :: Science/Research',
    'Intended Audience :: Developers',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development :: Libraries',
]

setup(
    name=DISTNAME,
    version=VERSION,
    license=LICENSE,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    classifiers=CLASSIFIERS,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    setup_requires=SETUP_REQUIRES,
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    url=URL,
    packages=find_packages(),
    include_package_data=True,
)



import numpy as np


nan = np.nan
r"NaN (Not a Number)"

inf = np.inf
r":math:`\infty` (infinite)"

pi = np.pi
r":math:`\pi`"

e = np.e
r":math:`e`"

euler_gamma = np.euler_gamma
r"Euler's :math:`\gamma`"



import pytest
import os.path
import numpy as np

from larray.tests.common import assert_array_equal, needs_pytables
from larray import Axis, IGroup, LGroup, LSet, ndtest, read_hdf
from larray.util.oset import OrderedSet

age = Axis('age=0..10')
lipro = Axis('lipro=P01..P05')
anonymous = Axis(range(3))
age_wildcard = Axis(10, 'wildcard')


def test_equals():
    a = Axis('a=a0..a2')
    # keys are single value
    assert a['a0'].equals(a['a0'])
    # keys are arrays
    assert a[np.array(['a0', 'a2'])].equals(a[np.array(['a0', 'a2'])])
    # axis names
    anonymous_axis = Axis('a0..a2')
    assert anonymous_axis['a0:a2'].equals(anonymous_axis['a0:a2'])
    assert not a['a0:a2'].equals(anonymous_axis['a0:a2'])
    assert not a['a0:a2'].equals(a.rename('other_name')['a0:a2'])
    # list of labels
    a02 = a['a0,a1,a2'] >> 'group'
    assert a02.equals(a02)
    a02_unamed = a['a0,a1,a2']
    assert not a02.equals(a02_unamed)
    a13 = a['a1,a2,a3'] >> 'group'
    assert not a02.equals(a13)
    a02_other_axis_name = a.rename('other_name')['a0,a1,a2'] >> 'group'
    assert not a02.equals(a02_other_axis_name)


# ################## #
#       LGroup       #
# ################## #

@pytest.fixture
def lgroups():
    class TestLGroup():
        def __init__(self):
            self.slice_both_named_wh_named_axis = LGroup('1:5', "full", age)
            self.slice_both_named = LGroup('1:5', "named")
            self.slice_both = LGroup('1:5')
            self.slice_start = LGroup('1:')
            self.slice_stop = LGroup(':5')
            self.slice_none_no_axis = LGroup(':')
            self.slice_none_wh_named_axis = LGroup(':', axis=lipro)
            self.slice_none_wh_anonymous_axis = LGroup(':', axis=anonymous)
            self.single_value = LGroup('P03')
            self.list = LGroup('P01,P03,P04')
            self.list_named = LGroup('P01,P03,P04', "P134")
    return TestLGroup()


def test_init_lgroup(lgroups):
    assert lgroups.slice_both_named_wh_named_axis.name == "full"
    assert lgroups.slice_both_named_wh_named_axis.key == slice(1, 5, None)
    assert lgroups.slice_both_named_wh_named_axis.axis is age
    assert lgroups.slice_both_named.name == "named"
    assert lgroups.slice_both_named.key == slice(1, 5, None)
    assert lgroups.slice_both.key == slice(1, 5, None)
    assert lgroups.slice_start.key == slice(1, None, None)
    assert lgroups.slice_stop.key == slice(None, 5, None)
    assert lgroups.slice_none_no_axis.key == slice(None, None, None)
    assert lgroups.slice_none_wh_named_axis.axis is lipro
    assert lgroups.slice_none_wh_anonymous_axis.axis is anonymous
    assert lgroups.single_value.key == 'P03'
    assert lgroups.list.key == ['P01', 'P03', 'P04']
    group = LGroup('1:5', age, age)
    assert group.name == age.name
    group = age['1:5'] >> age
    assert group.name == age.name
    group2 = LGroup('1', axis=age)
    group = LGroup('1', group2, axis=age)
    assert group.name == '1'
    group = age['1'] >> group2
    assert group.name == '1'
    group2 = LGroup('1:5', 'age', age)
    group = LGroup('1:5', group2, axis=age)
    assert group.name == group2.name
    group = age['1:5'] >> group2
    assert group.name == group2.name
    axis = Axis('axis=a,a0..a3,b,b0..b3,c,c0..c3')
    for code in axis.matching(regex='^.$'):
        group = axis.startingwith(code) >> code
        assert group.equals(axis.startingwith(code) >> str(code))


def test_eq_lgroup(lgroups):
    # with axis vs no axis do not compare equal
    # lgroups.slice_both == lgroups.slice_both_named_wh_named_axis
    assert lgroups.slice_both == lgroups.slice_both_named
    res = lgroups.slice_both_named_wh_named_axis == age[1:5]
    assert isinstance(res, np.ndarray)
    assert res.shape == (5,)
    assert res.all()
    assert lgroups.slice_both == LGroup(slice(1, 5))
    assert lgroups.slice_start == LGroup(slice(1, None))
    assert lgroups.slice_stop == LGroup(slice(5))
    assert lgroups.slice_none_no_axis == LGroup(slice(None))
    assert lgroups.single_value == LGroup('P03')
    assert lgroups.list == LGroup(['P01', 'P03', 'P04'])
    assert lgroups.list_named == LGroup(['P01', 'P03', 'P04'])
    assert lgroups.slice_both == slice(1, 5)
    assert lgroups.slice_start == slice(1, None)
    assert lgroups.slice_stop == slice(5)
    assert lgroups.slice_none_no_axis == slice(None)
    assert lgroups.single_value == 'P03'
    assert lgroups.list == ['P01', 'P03', 'P04']
    assert lgroups.list_named == ['P01', 'P03', 'P04']


def test_getitem_lgroup():
    axis = Axis("a=a0,a1")
    assert axis['a0'][0] == 'a'
    assert axis['a0'][1] == '0'
    assert axis['a0':'a1'][1] == 'a1'
    assert axis[:][1] == 'a1'
    assert list(axis[:][0:2]) == ['a0', 'a1']
    assert list((axis[:][[1, 0]])) == ['a1', 'a0']
    assert axis[['a0', 'a1', 'a0']][2] == 'a0'
    assert axis[('a0', 'a1', 'a0')][2] == 'a0'
    assert axis[ndtest("a=a0,a1,a0")][2] == 2


def test_sorted_lgroup():
    assert sorted(LGroup(['c', 'd', 'a', 'b'])) == [LGroup('a'), LGroup('b'), LGroup('c'), LGroup('d')]


def test_asarray_lgroup(lgroups):
    assert_array_equal(np.asarray(lgroups.slice_both_named_wh_named_axis), np.array([1, 2, 3, 4, 5]))
    assert_array_equal(np.asarray(lgroups.slice_none_wh_named_axis), np.array(['P01', 'P02', 'P03', 'P04', 'P05']))


def test_hash_lgroup(lgroups):
    # this test is a lot less important than what it used to, because we cannot have Group ticks on an axis anymore
    d = {lgroups.slice_both: 1, lgroups.single_value: 2, lgroups.list_named: 3}
    assert d.get(lgroups.slice_both) == 1
    assert d.get(lgroups.single_value) == 2
    assert d.get(lgroups.list) == 3
    assert d.get(lgroups.list_named) == 3


def test_repr_lgroup(lgroups):
    assert repr(lgroups.slice_both_named_wh_named_axis) == "age[1:5] >> 'full'"
    assert repr(lgroups.slice_both_named) == "LGroup(slice(1, 5, None)) >> 'named'"
    assert repr(lgroups.slice_both) == "LGroup(slice(1, 5, None))"
    assert repr(lgroups.list) == "LGroup(['P01', 'P03', 'P04'])"
    assert repr(lgroups.slice_none_no_axis) == "LGroup(slice(None, None, None))"
    assert repr(lgroups.slice_none_wh_named_axis) == "lipro[:]"
    assert repr(lgroups.slice_none_wh_anonymous_axis) == "LGroup(slice(None, None, None), axis=Axis([0, 1, 2], None))"


def test_to_int_lgroup():
    a = Axis(['42'], 'a')
    assert int(a['42']) == 42


def test_to_float_lgroup():
    a = Axis(['42'], 'a')
    assert float(a['42']) == 42.0


def test_to_lset_lgroup():
    alpha = Axis('alpha=a,b,c,d')

    # list key
    lg = LGroup(['a', 'a', 'c'], axis=alpha)
    res = lg.set()
    assert isinstance(res, LSet)
    assert res.axis is alpha
    assert res.key == OrderedSet(['a', 'c'])

    # scalar key
    lg = LGroup('c', axis=alpha)
    res = lg.set()
    assert isinstance(res, LSet)
    assert res.axis is alpha
    assert res.key == OrderedSet(['c'])


@needs_pytables
def test_h5_io_lgroup(tmpdir):
    fpath = os.path.join(str(tmpdir), 'lgroups.h5')
    age.to_hdf(fpath)

    named = age[':5'] >> 'age_05'
    named_axis_not_in_file = lipro['P01,P03,P05'] >> 'P_odd'
    anonymous = age[':5']
    wildcard = age_wildcard[':5'] >> 'age_w_05'
    string_group = Axis(['@!àéè&%µ$~', '/*-+_§()><', 'another label'], 'string_axis')[:] >> 'string_group'

    # ---- default behavior ----
    # named group
    named.to_hdf(fpath)
    named2 = read_hdf(fpath, key=named.name)
    assert all(named == named2)
    # anonymous group
    with pytest.raises(ValueError, match="Argument key must be provided explicitly in case of anonymous group"):
        anonymous.to_hdf(fpath)
    # wildcard group
    wildcard.to_hdf(fpath)
    wildcard2 = read_hdf(fpath, key=wildcard.name)
    assert all(wildcard == wildcard2)
    # associated axis not saved yet
    named_axis_not_in_file.to_hdf(fpath)
    named2 = read_hdf(fpath, key=named_axis_not_in_file.name)
    assert all(named_axis_not_in_file == named2)
    # string group
    string_group.to_hdf(fpath)
    string_group2 = read_hdf(fpath, key=string_group.name)
    assert all(string_group == string_group2)

    # ---- specific hdf group + key ----
    hdf_group = 'my_groups'
    # named group
    key = hdf_group + '/named_group'
    named.to_hdf(fpath, key)
    named2 = read_hdf(fpath, key=key)
    assert all(named == named2)
    # anonymous group
    key = hdf_group + '/anonymous_group'
    anonymous.to_hdf(fpath, key)
    anonymous2 = read_hdf(fpath, key=key)
    assert anonymous2.name is None
    assert all(anonymous == anonymous2)
    # wildcard group
    key = hdf_group + '/wildcard_group'
    wildcard.to_hdf(fpath, key)
    wildcard2 = read_hdf(fpath, key=key)
    assert all(wildcard == wildcard2)
    # associated axis not saved yet
    key = hdf_group + '/named_group_axis_not_in_file'
    named_axis_not_in_file.to_hdf(fpath, key=key)
    named2 = read_hdf(fpath, key=key)
    assert all(named_axis_not_in_file == named2)

    # ---- specific axis_key ----
    axis_key = 'axes/associated_axis_0'
    # named group
    named.to_hdf(fpath, axis_key=axis_key)
    named2 = read_hdf(fpath, key=named.name)
    assert all(named == named2)
    # anonymous group
    key = 'anonymous'
    anonymous.to_hdf(fpath, key=key, axis_key=axis_key)
    anonymous2 = read_hdf(fpath, key=key)
    assert anonymous2.name is None
    assert all(anonymous == anonymous2)
    # wildcard group
    wildcard.to_hdf(fpath, axis_key=axis_key)
    wildcard2 = read_hdf(fpath, key=wildcard.name)
    assert all(wildcard == wildcard2)
    # associated axis not saved yet
    axis_key = 'axes/associated_axis_1'
    named_axis_not_in_file.to_hdf(fpath, axis_key=axis_key)
    named2 = read_hdf(fpath, key=named_axis_not_in_file.name)
    assert all(named_axis_not_in_file == named2)


# ################## #
#        LSet        #
# ################## #

def test_or_lset():
    # without axis
    assert LSet(['a', 'b']) | LSet(['c', 'd']) == LSet(['a', 'b', 'c', 'd'])
    assert LSet(['a', 'b', 'c']) | LSet(['c', 'd']) == LSet(['a', 'b', 'c', 'd'])
    alpha = Axis('alpha=a,b,c,d')
    res = alpha['a', 'b'].set() | alpha['c', 'd'].set()
    assert res.axis is alpha
    assert res == alpha['a', 'b', 'c', 'd'].set()
    assert alpha['a', 'b', 'c'].set() | alpha['c', 'd'].set() == alpha['a', 'b', 'c', 'd'].set()
    alpha = Axis('alpha=a,b,c,d')
    res = alpha['a', 'b'].set() | alpha['c', 'd']
    assert res.axis is alpha
    assert res == alpha['a', 'b', 'c', 'd'].set()
    assert alpha['a', 'b', 'c'].set() | alpha['c', 'd'] == alpha['a', 'b', 'c', 'd'].set()
    alpha = Axis('alpha=a,b,c,d')
    res = alpha['a', 'b'].set().named('ab') | alpha['c', 'd'].set().named('cd')
    assert res.axis is alpha
    assert res.name == 'ab | cd'
    assert res == alpha['a', 'b', 'c', 'd'].set()
    assert alpha['a', 'b', 'c'].set() | alpha['c', 'd'] == alpha['a', 'b', 'c', 'd'].set()
    num = Axis(range(10), 'num')
    assert num[1, 5, 3].set() | 4 == num[1, 5, 3, 4].set()
    assert num[1, 5, 3].set() | num[4] == num[1, 5, 3, 4].set()
    assert num[4].set() | num[1, 5, 3] == num[4, 1, 5, 3].set()
    assert num[:2].set() | num[8:] == num[0, 1, 2, 8, 9].set()
    assert num[:2].set() | num[5] == num[0, 1, 2, 5].set()


def test_and_lset():
    # without axis
    assert LSet(['a', 'b', 'c']) & LSet(['c', 'd']) == LSet(['c'])
    alpha = Axis('alpha=a,b,c,d')
    res = alpha['a', 'b', 'c'].named('abc').set() & alpha['c', 'd'].named('cd')
    assert res.axis is alpha
    assert res.name == 'abc & cd'
    assert res == alpha[['c']].set()


def test_sub_lset():
    assert LSet(['a', 'b', 'c']) - LSet(['c', 'd']) == LSet(['a', 'b'])
    assert LSet(['a', 'b', 'c']) - ['c', 'd'] == LSet(['a', 'b'])
    assert LSet(['a', 'b', 'c']) - 'b' == LSet(['a', 'c'])
    assert LSet([1, 2, 3]) - 4 == LSet([1, 2, 3])
    assert LSet([1, 2, 3]) - 2 == LSet([1, 3])


# ################## #
#       IGroup       #
# ################## #

@pytest.fixture
def igroups():
    class TestIGroup():
        def __init__(self):
            self.code_axis = Axis('code=a0..a4')
            self.slice_both_named = self.code_axis.i[1:4] >> 'a123'
            self.slice_both = self.code_axis.i[1:4]
            self.slice_start = self.code_axis.i[1:]
            self.slice_stop = self.code_axis.i[:4]
            self.slice_none = self.code_axis.i[:]
            self.first_value = self.code_axis.i[0]
            self.last_value = self.code_axis.i[-1]
            self.list = self.code_axis.i[[0, 1, -2, -1]]
            self.tuple = self.code_axis.i[0, 1, -2, -1]
    return TestIGroup()


def _assert_array_equal_is_true_array(a, b):
    res = a == b
    assert isinstance(res, np.ndarray)
    assert res.shape == np.asarray(b).shape
    assert res.all()


def test_asarray_igroup(igroups):
    assert_array_equal(np.asarray(igroups.slice_both), np.array(['a1', 'a2', 'a3']))


def test_eq_igroup(igroups):
    _assert_array_equal_is_true_array(igroups.slice_both, ['a1', 'a2', 'a3'])
    _assert_array_equal_is_true_array(igroups.slice_both_named, ['a1', 'a2', 'a3'])
    _assert_array_equal_is_true_array(igroups.slice_both, igroups.slice_both_named)
    _assert_array_equal_is_true_array(igroups.slice_both_named, igroups.slice_both)
    _assert_array_equal_is_true_array(igroups.slice_start, ['a1', 'a2', 'a3', 'a4'])
    _assert_array_equal_is_true_array(igroups.slice_stop, ['a0', 'a1', 'a2', 'a3'])
    _assert_array_equal_is_true_array(igroups.slice_none, ['a0', 'a1', 'a2', 'a3', 'a4'])
    assert igroups.first_value == 'a0'
    assert igroups.last_value == 'a4'
    _assert_array_equal_is_true_array(igroups.list, ['a0', 'a1', 'a3', 'a4'])
    _assert_array_equal_is_true_array(igroups.tuple, ['a0', 'a1', 'a3', 'a4'])


def test_getitem_igroup():
    axis = Axis("a=a0,a1")
    assert axis.i[0][0] == 'a'
    assert axis.i[0][1] == '0'
    assert axis.i[0:1][1] == 'a1'
    assert axis.i[:][1] == 'a1'
    assert list(axis.i[:][0:2]) == ['a0', 'a1']
    assert list((axis.i[:][[1, 0]])) == ['a1', 'a0']
    assert axis.i[[0, 1, 0]][2] == 'a0'
    assert axis.i[(0, 1, 0)][2] == 'a0'


def test_getattr_igroup():
    agg = Axis(['a1:a2', ':a2', 'a1:'], 'agg')
    assert agg.i[0].split(':') == ['a1', 'a2']
    assert agg.i[1].split(':') == ['', 'a2']
    assert agg.i[2].split(':') == ['a1', '']


def test_dir_igroup():
    agg = Axis(['a', 1], 'agg')
    assert 'split' in dir(agg.i[0])
    assert 'strip' in dir(agg.i[0])
    assert 'strip' in dir(agg.i[0])


def test_repr_igroup(igroups):
    assert repr(igroups.slice_both_named) == "code.i[1:4] >> 'a123'"
    assert repr(igroups.slice_both) == "code.i[1:4]"
    assert repr(igroups.slice_start) == "code.i[1:]"
    assert repr(igroups.slice_stop) == "code.i[:4]"
    assert repr(igroups.slice_none) == "code.i[:]"
    assert repr(igroups.first_value) == "code.i[0]"
    assert repr(igroups.last_value) == "code.i[-1]"
    assert repr(igroups.list) == "code.i[0, 1, -2, -1]"
    assert repr(igroups.tuple) == "code.i[0, 1, -2, -1]"


def test_to_int_igroup():
    a = Axis(['42'], 'a')
    assert int(a.i[0]) == 42


def test_to_float_igroup():
    a = Axis(['42'], 'a')
    assert float(a.i[0]) == 42.0


def test_to_lset_igroup():
    alpha = Axis('alpha=a,b,c,d')

    # list key
    ig = IGroup([0, 0, 2], axis=alpha)
    res = ig.set()
    assert isinstance(res, LSet)
    assert res.axis is alpha
    assert res.key == OrderedSet(['a', 'c'])

    # scalar key
    ig = IGroup(2, axis=alpha)
    res = ig.set()
    assert isinstance(res, LSet)
    assert res.axis is alpha
    assert res.key == OrderedSet(['c'])


@needs_pytables
def test_h5_io_igroup(tmpdir):
    fpath = os.path.join(str(tmpdir), 'igroups.h5')
    age.to_hdf(fpath)

    named = age.i[:6] >> 'age_05'
    named_axis_not_in_file = lipro.i[1::2] >> 'P_odd'
    anonymous = age.i[:6]
    wildcard = age_wildcard.i[:6] >> 'age_w_05'

    # ---- default behavior ----
    # named group
    named.to_hdf(fpath)
    named2 = read_hdf(fpath, key=named.name)
    assert all(named == named2)
    # anonymous group
    with pytest.raises(ValueError, match="Argument key must be provided explicitly in case of anonymous group"):
        anonymous.to_hdf(fpath)
    # wildcard group
    wildcard.to_hdf(fpath)
    wildcard2 = read_hdf(fpath, key=wildcard.name)
    assert all(wildcard == wildcard2)
    # associated axis not saved yet
    named_axis_not_in_file.to_hdf(fpath)
    named2 = read_hdf(fpath, key=named_axis_not_in_file.name)
    assert all(named_axis_not_in_file == named2)

    # ---- specific hdf group + key ----
    hdf_group = 'my_groups'
    # named group
    key = hdf_group + '/named_group'
    named.to_hdf(fpath, key)
    named2 = read_hdf(fpath, key=key)
    assert all(named == named2)
    # anonymous group
    key = hdf_group + '/anonymous_group'
    anonymous.to_hdf(fpath, key)
    anonymous2 = read_hdf(fpath, key=key)
    assert anonymous2.name is None
    assert all(anonymous == anonymous2)
    # wildcard group
    key = hdf_group + '/wildcard_group'
    wildcard.to_hdf(fpath, key)
    wildcard2 = read_hdf(fpath, key=key)
    assert all(wildcard == wildcard2)
    # associated axis not saved yet
    key = hdf_group + '/named_group_axis_not_in_file'
    named_axis_not_in_file.to_hdf(fpath, key=key)
    named2 = read_hdf(fpath, key=key)
    assert all(named_axis_not_in_file == named2)

    # ---- specific axis_key ----
    axis_key = 'axes/associated_axis_0'
    # named group
    named.to_hdf(fpath, axis_key=axis_key)
    named2 = read_hdf(fpath, key=named.name)
    assert all(named == named2)
    # anonymous group
    key = 'anonymous'
    anonymous.to_hdf(fpath, key=key, axis_key=axis_key)
    anonymous2 = read_hdf(fpath, key=key)
    assert anonymous2.name is None
    assert all(anonymous == anonymous2)
    # wildcard group
    wildcard.to_hdf(fpath, axis_key=axis_key)
    wildcard2 = read_hdf(fpath, key=wildcard.name)
    assert all(wildcard == wildcard2)
    # associated axis not saved yet
    axis_key = 'axes/associated_axis_1'
    named_axis_not_in_file.to_hdf(fpath, axis_key=axis_key)
    named2 = read_hdf(fpath, key=named_axis_not_in_file.name)
    assert all(named_axis_not_in_file == named2)


if __name__ == "__main__":
    pytest.main()



"""
Misc tools
"""

import __main__
import math
import itertools
import os
import sys
import operator
import warnings
from textwrap import wrap
from functools import reduce, wraps
from itertools import product
from collections import defaultdict

import numpy as np
import pandas as pd

try:
    np.set_printoptions(legacy='1.13')
except TypeError:
    pass


def is_interactive_interpreter():
    try:
        # When running using IPython, sys.ps1 is always defined, so we cannot use the standard "hasattr(sys, 'ps1')"
        # Additionally, an InProcessInteractiveShell can have a __main__ module with a file
        main_lacks_file = not hasattr(__main__, '__file__')
        return main_lacks_file or get_ipython().__class__.__name__ == 'InProcessInteractiveShell'
    except NameError:
        return hasattr(sys, 'ps1')


def prod(values):
    return reduce(operator.mul, values, 1)


def format_value(value, missing, precision=None):
    if isinstance(value, float):
        # nans print as "-1.#J", let's use something nicer
        if value != value:
            return missing
        elif precision is not None:
            return f'{value:.{precision}f}'
        else:
            return str(value)
    elif isinstance(value, np.ndarray) and value.shape:
        # prevent numpy's default wrapping
        return str(list(value)).replace(',', '')
    else:
        return str(value)


def get_col_width(table, index):
    return max(len(row[index]) for row in table)


def longest_word(s):
    """Return length of the longest word in the given string

    Parameters
    ----------
    s : str
        string to check
    Returns
    -------
    int
        length of longest word

    Examples
    --------
    >>> longest_word('12 123 1234')
    4
    >>> longest_word('12 1234 123')
    4
    >>> longest_word('123 12 123')
    3
    >>> longest_word('')
    0
    >>> longest_word(' ')
    0
    """
    return max(len(w) for w in s.split()) if s and not s.isspace() else 0


def get_min_width(table, index):
    return max(longest_word(row[index]) for row in table)


def table2str(table, missing, summarize=True, maxwidth=200, numedges='auto', sep='  ', cont='...',
              keepcols=0, precision=None):
    """
    table is a list of lists
    :type table: list of list
    """
    if not table:
        return ''
    numcol = max(len(row) for row in table)
    # pad rows that have too few columns
    for row in table:
        if len(row) < numcol:
            row.extend([''] * (numcol - len(row)))
    formatted = [[format_value(value, missing, precision) for value in row]
                 for row in table]
    maxwidths = [get_col_width(formatted, i) for i in range(numcol)]

    total_colwidth = sum(maxwidths)
    sep_width = (numcol - 1) * len(sep)
    if total_colwidth + sep_width > maxwidth:
        minwidths = [get_min_width(formatted, i) for i in range(numcol)]
        available_width = maxwidth - sep_width - sum(minwidths)
        if available_width >= 0:
            ratio = available_width / total_colwidth
            colwidths = [minw + int(maxw * ratio)
                         for minw, maxw in zip(minwidths, maxwidths)]
        else:
            # need to exceed maxwidth or hide some data
            if summarize:
                if numedges == 'auto':
                    w = sum(minwidths[:keepcols]) + len(cont)
                    maxedges = (numcol - keepcols) // 2
                    if maxedges:
                        maxi = 0
                        for i in range(1, maxedges + 1):
                            w += minwidths[i] + minwidths[-i]
                            # + 1 for the "continuation" column
                            ncol = keepcols + i * 2 + 1
                            sepw = (ncol - 1) * len(sep)
                            maxi = i
                            if w + sepw > maxwidth:
                                break
                        numedges = maxi - 1
                    else:
                        numedges = 0
                head = keepcols + numedges
                tail = -numedges if numedges else numcol
                formatted = [row[:head] + [cont] + row[tail:]
                             for row in formatted]
                colwidths = minwidths[:head] + [len(cont)] + minwidths[tail:]
            else:
                colwidths = minwidths
    else:
        colwidths = maxwidths

    lines = []
    for row in formatted:
        wrapped_row = [wrap(value, width) if width > 0 else value
                       for value, width in zip(row, colwidths)]
        maxlines = max(len(value) for value in wrapped_row)
        newlines = [[] for _ in range(maxlines)]
        for value, width in zip(wrapped_row, colwidths):
            for i in range(maxlines):
                chunk = value[i] if i < len(value) else ''
                newlines[i].append(chunk.rjust(width))
        lines.extend(newlines)
    return '\n'.join(sep.join(row) for row in lines)


# copied from itertools recipes
def unique(iterable):
    """
    Yields all elements once, preserving order. Remember all elements ever seen.
    >>> list(unique('AAAABBBCCDAABBB'))
    ['A', 'B', 'C', 'D']
    """
    seen = set()
    seen_add = seen.add
    for element in iterable:
        if element not in seen:
            seen_add(element)
            yield element


def unique_list(iterable, res=None, seen=None):
    """
    Returns a list of all unique elements, preserving order. Remember all elements ever seen.
    >>> unique_list('AAAABBBCCDAABBB')
    ['A', 'B', 'C', 'D']
    """
    if res is None:
        res = []
    res_append = res.append
    if seen is None:
        seen = set()
    seen_add = seen.add
    for element in iterable:
        if element not in seen:
            seen_add(element)
            res_append(element)
    return res


def unique_multi(iterable_of_iterables):
    """
    Returns a list of all unique elements across multiple iterables. Elements of earlier iterables will come first.
    """
    seen = set()
    res = []
    for iterable in iterable_of_iterables:
        unique_list(iterable, res, seen)
    return res


def has_duplicates(iterable):
    """
    Returns whether or not iterable contains any duplicated element.
    """
    # using a dict is faster than using a set (at least for Python <= 3.9)
    seen = {}
    for element in iterable:
        if element in seen:
            return True
        else:
            seen[element] = True
    return False


def duplicates(iterable):
    """
    List duplicated elements once, preserving order. Remember all elements ever seen.
    """
    # duplicates('AAAABBBCCDAABBB') --> A B C
    counts = defaultdict(int)
    for element in iterable:
        counts[element] += 1
        if counts[element] == 2:
            yield element


def rproduct(*i):
    return product(*[x[::-1] for x in i])


def light_product(*iterables, **kwargs):
    """Cartesian product of input iterables, replacing repeated values by empty strings.

    Parameters
    ----------
    *iterables : iterable
        Input iterables
    repeat : int, optional
        Number of times to repeat (reuse) input iterables

    Returns
    -------
    Generator

    Examples
    --------
    >>> list(light_product('ab', range(3)))
    [('a', 0), ('', 1), ('', 2), ('b', 0), ('', 1), ('', 2)]
    >>> list(light_product('ab', repeat=2))
    [('a', 'a'), ('', 'b'), ('b', 'a'), ('', 'b')]
    """
    repeat = kwargs.pop('repeat', 1)
    p = product(*iterables, repeat=repeat)
    prev_t = (None,) * len(iterables) * repeat
    for t in p:
        yield tuple(e if e != prev_e else ''
                    for e, prev_e in zip(t, prev_t))
        prev_t = t


def array_nan_equal(a, b):
    if np.issubdtype(a.dtype, str) and np.issubdtype(b.dtype, str):
        return np.array_equal(a, b)
    else:
        return np.all((a == b) | (np.isnan(a) & np.isnan(b)))


def unzip(iterable):
    return list(zip(*iterable))


class ReprString(str):
    def __repr__(self):
        return self


def array_lookup(array, mapping):
    """pass all elements of an np.ndarray through a mapping"""
    array = np.asarray(array)
    # TODO: this must be cached in the Axis
    # TODO: range axes should be optimized (reuse Pandas 0.18 indexes)
    sorted_keys, sorted_values = tuple(zip(*sorted(mapping.items())))
    sorted_keys = np.array(sorted_keys)
    # prevent an array of booleans from matching a integer axis (sorted_keys)
    # XXX: we might want to allow signed and unsigned integers to match against each other
    if array.dtype.kind != sorted_keys.dtype.kind:
        raise KeyError('key has not the same dtype than axis')
    # TODO: it is very important to fail quickly, so guess_axis should try this in chunks
    # (first test first element of key, if several axes match, try [1:11] elements, [12:112], [113:1113], ...
    if not np.all(np.in1d(array, sorted_keys)):
        raise KeyError('all keys not in array')

    sorted_values = np.array(sorted_values)
    if not len(array):
        return np.empty(0, dtype=sorted_values.dtype)
    indices = np.searchsorted(sorted_keys, array)
    return sorted_values[indices]


def array_lookup2(array, sorted_keys, sorted_values):
    """pass all elements of an np.ndarray through a "mapping" """
    if not len(array):
        return np.empty(0, dtype=sorted_values.dtype)

    array = np.asarray(array)
    # TODO: range axes should be optimized (reuse Pandas 0.18 indexes)

    # prevent an array of booleans from matching a integer axis (sorted_keys)
    # XXX: we might want to allow signed and unsigned integers to match against each other
    if array.dtype.kind != sorted_keys.dtype.kind:
        raise KeyError('key has not the same dtype than axis')
    # TODO: it is very important to fail quickly, so guess_axis should try this in chunks
    # (first test first element of key, if several axes match, try [1:11] elements, [12:112], [113:1113], ...
    if not np.all(np.in1d(array, sorted_keys)):
        raise KeyError('all keys not in array')

    indices = np.searchsorted(sorted_keys, array)
    return sorted_values[indices]


def split_on_condition(seq, condition):
    """splits an iterable into two lists depending on a condition

    Parameters
    ----------
    seq : iterable
    condition : function(e) -> True or False

    Returns
    -------
    a, b: list

    Notes
    -----
    If the condition can be inlined into a list comprehension, a double list comprehension is faster than this function.
    So if performance is crucial, you should inline this function with the condition itself inlined.
    """
    a, b = [], []
    append_a, append_b = a.append, b.append
    for e in seq:
        append_a(e) if condition(e) else append_b(e)
    return a, b


def split_on_values(seq, values):
    """splits an iterable into two lists depending on a list of values

    Parameters
    ----------
    seq : iterable
    values : iterable
        set of values which must go to the first list

    Returns
    -------
    a, b: list
    """
    values = set(values)
    a, b = [], []
    append_a, append_b = a.append, b.append
    for e in seq:
        append_a(e) if e in values else append_b(e)
    return a, b


def skip_comment_cells(lines):
    def notacomment(v):
        return not v.startswith('#')
    for line in lines:
        stripped_line = list(itertools.takewhile(notacomment, line))
        if stripped_line:
            yield stripped_line


def strip_rows(lines):
    """
    returns an iterator of lines with trailing blank (empty or
    which contain only space) cells.
    """
    def isblank(s):
        return s == '' or s.isspace()
    for line in lines:
        rev_line = list(itertools.dropwhile(isblank, reversed(line)))
        yield list(reversed(rev_line))


def size2str(value):
    """
    >>> size2str(0)
    '0 bytes'
    >>> size2str(100)
    '100 bytes'
    >>> size2str(1023)
    '1023 bytes'
    >>> size2str(1024)
    '1.00 Kb'
    >>> size2str(2000)
    '1.95 Kb'
    >>> size2str(10000000)
    '9.54 Mb'
    >>> size2str(1.27 * 1024 ** 3)
    '1.27 Gb'
    """
    units = ["bytes", "Kb", "Mb", "Gb", "Tb", "Pb"]
    scale = int(math.log(value, 1024)) if value else 0
    scaled_value = value / 1024.0 ** scale
    return f"{scaled_value:.2f} {units[scale]}" if scale > 0 else f"{int(scaled_value)} bytes"


def find_closing_chr(s, start=0):
    """

    Parameters
    ----------
    s : str
        string to search the characters. s[start] must be in '({['
    start : int, optional
        position in the string from which to start searching

    Returns
    -------
    position of matching brace

    Examples
    --------
    >>> find_closing_chr('(a) + (b)')
    2
    >>> find_closing_chr('(a) + (b)', 6)
    8
    >>> find_closing_chr('(a{b[c(d)e]f}g)')
    14
    >>> find_closing_chr('(a{b[c(d)e]f}g)', 2)
    12
    >>> find_closing_chr('(a{b[c(d)e]f}g)', 4)
    10
    >>> find_closing_chr('(a{b[c(d)e]f}g)', 6)
    8
    >>> find_closing_chr('((a) + (b))')
    10
    >>> find_closing_chr('((a) + (b))')
    10
    >>> find_closing_chr('((a) + (b))')
    10
    >>> find_closing_chr('({)}')
    Traceback (most recent call last):
      ...
    ValueError: malformed expression: expected '}' but found ')'
    >>> find_closing_chr('({}})')
    Traceback (most recent call last):
      ...
    ValueError: malformed expression: expected ')' but found '}'
    >>> find_closing_chr('}()')
    Traceback (most recent call last):
      ...
    ValueError: malformed expression: found '}' before '{'
    >>> find_closing_chr('(()')
    Traceback (most recent call last):
      ...
    ValueError: malformed expression: reached end of string without finding the expected ')'
    """
    opening, closing = '({[', ')}]'
    match = {o: c for o, c in zip(opening, closing)}
    match.update({c: o for o, c in zip(opening, closing)})
    opening_set, closing_set = set(opening), set(closing)

    needle = s[start]
    assert needle in match
    last_open = []
    for pos in range(start, len(s)):
        c = s[pos]
        if c in match:
            if c in opening_set:
                last_open.append(c)
            if c in closing_set:
                if not last_open:
                    raise ValueError(f"malformed expression: found '{c}' before '{match[c]}'")
                expected = match[last_open.pop()]
                if c != expected:
                    raise ValueError(f"malformed expression: expected '{expected}' but found '{c}'")
                if not last_open:
                    assert c == match[needle]
                    return pos
    raise ValueError(f"malformed expression: reached end of string without finding the expected '{match[needle]}'")


def float_error_handler_factory(stacklevel):
    def error_handler(error, flag):
        if error == 'invalid value':
            error = 'invalid value (NaN)'
            extra = ' (this is typically caused by a 0 / 0)'
        else:
            extra = ''
        warnings.warn(f"{error} encountered during operation{extra}", RuntimeWarning, stacklevel=stacklevel)
    return error_handler


def _isintstring(s):
    """
    Return True if the passed string represents an integer.
    Zero padded integers are considered as strings and not integers.

    Parameters
    ----------
    s : str
        string to test if representing an integer.

    Examples
    --------
    >>> _isintstring('12')
    True
    >>> _isintstring('-12')
    True
    >>> _isintstring('a1')
    False
    >>> _isintstring('01')
    False
    """
    def isposint(s):
        # exclude zero padded strings
        return s.isdigit() and not (len(s) > 1 and s[0] == '0')
    return isposint(s) or (len(s) > 1 and s[0] == '-' and isposint(s[1:]))


def _parse_bound(s, stack_depth=1, parse_int=True):
    """Parse a string representing a single value, converting int-like strings to integers and evaluating expressions
    within {}.

    Parameters
    ----------
    s : str
        string to evaluate
    stack_depth : int
        how deep to go in the stack to get local variables for evaluating {expressions}.

    Returns
    -------
    any

    Examples
    --------
    >>> _parse_bound(' a ')
    'a'
    >>> # returns None
    >>> _parse_bound(' ')
    >>> ext = 1
    >>> _parse_bound(' {ext + 1} ')
    2
    >>> _parse_bound('42')
    42
    >>> _parse_bound('01')
    '01'
    """
    s = s.strip()
    if s == '':
        return None
    elif s[0] == '{':
        expr = s[1:find_closing_chr(s)]
        return eval(expr, sys._getframe(stack_depth).f_locals)
    elif parse_int and _isintstring(s):
        return int(s)
    else:
        return s


def _isnoneslice(v):
    """
    Checks if input is slice(None) object.
    """
    return isinstance(v, slice) and v.start is None and v.stop is None and v.step is None


def _seq_summary(seq, n=3, repr_func=repr, sep=' '):
    """
    Returns a string representing a sequence by showing only the n first and last elements.

    Examples
    --------
    >>> _seq_summary(range(10), 2)
    '0 1 ... 8 9'
    """
    if len(seq) <= 2 * n:
        short_seq = [repr_func(v) for v in seq]
    else:
        short_seq = [repr_func(v) for v in seq[:n]] + ['...'] + [repr_func(v) for v in seq[-n:]]
    return sep.join(short_seq)


def index_by_id(seq, value):
    """
    Returns position of an object in a sequence.

    Raises an error if the object is not in the list.

    Parameters
    ----------
    seq : sequence
        Any sequence (list, tuple, str, unicode).

    value : object
        Object for which you want to retrieve its position in the sequence.

    Raises
    ------
    ValueError
        If `value` object is not contained in the sequence.

    Examples
    --------
    >>> from larray import Axis
    >>> age = Axis('age=0..9')
    >>> sex = Axis('sex=M,F')
    >>> time = Axis('time=2007..2010')
    >>> index_by_id([age, sex, time], sex)
    1
    >>> gender = Axis('sex=M,F')
    >>> index_by_id([age, sex, time], gender)
    Traceback (most recent call last):
        ...
    ValueError: sex is not in list
    >>> gender = sex
    >>> index_by_id([age, sex, time], gender)
    1
    """
    for i, item in enumerate(seq):
        if item is value:
            return i
    raise ValueError(f"{value} is not in list")


def renamed_to(newfunc, old_name, stacklevel=2):
    def wrapper(*args, **kwargs):
        warnings.warn(f"{old_name}() is deprecated. Use {newfunc.__name__}() instead.",
                      FutureWarning, stacklevel=stacklevel)
        return newfunc(*args, **kwargs)
    return wrapper


# deprecate_kwarg is derived from pandas.util._decorators (0.21)
def deprecate_kwarg(old_arg_name, new_arg_name, mapping=None, arg_converter=None, stacklevel=2):
    if mapping is not None and not isinstance(mapping, dict):
        raise TypeError("mapping from old to new argument values must be dict!")

    def _deprecate_kwarg(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            old_arg_value = kwargs.pop(old_arg_name, None)
            if old_arg_value is not None:
                if mapping is not None:
                    new_arg_value = mapping.get(old_arg_value, old_arg_value)
                elif arg_converter is not None:
                    new_arg_value = arg_converter(old_arg_value)
                else:
                    new_arg_value = old_arg_value
                msg = "The {old_name}={old_val!r} keyword is deprecated, use {new_name}={new_val!r} instead"\
                    .format(old_name=old_arg_name, old_val=old_arg_value, new_name=new_arg_name,
                            new_val=new_arg_value)

                warnings.warn(msg, FutureWarning, stacklevel=stacklevel)
                if new_arg_name in kwargs:
                    msg = "Can only specify '{old_name}' or '{new_name}', not both"\
                        .format(old_name=old_arg_name, new_name=new_arg_name)
                    raise ValueError(msg)
                else:
                    kwargs[new_arg_name] = new_arg_value
            return func(*args, **kwargs)
        return wrapper
    return _deprecate_kwarg


class lazy_attribute:
    """
    Decorate a method of a class and turn it into an attribute on the instance
    when first called. Should obviously only be used when the result of the method is constant.
    """
    def __init__(self, func):
        self.func = func
        self.__doc__ = func.__doc__

    # descriptor protocol
    # see https://docs.python.org/3/reference/datamodel.html#implementing-descriptors
    def __get__(self, instance, owner):
        if instance is None:
            return self

        func = self.func
        value = func(instance)
        setattr(instance, func.__name__, value)
        return value


def inverseop(opname):
    comparison_ops = {
        'lt': 'gt',
        'gt': 'lt',
        'le': 'ge',
        'ge': 'le',
        'eq': 'eq',
        'ne': 'ne'
    }
    if opname in comparison_ops:
        return comparison_ops[opname]
    else:
        return 'r' + opname


_numeric_kinds = 'buifc'    # Boolean, Unsigned integer, Integer, Float, Complex
_string_kinds = 'SU'        # String, Unicode
_meta_kind = {k: 'str' for k in _string_kinds}
_meta_kind.update({k: 'numeric' for k in _numeric_kinds})


def common_type(arrays):
    """
    Returns a type which is common to the input arrays.
    All input arrays can be safely cast to the returned dtype without loss of information.

    Notes
    -----
    If list of arrays mixes 'numeric' and 'string' types, the function returns 'object' as common type.
    """
    arrays = [np.asarray(a) for a in arrays]
    dtypes = [a.dtype for a in arrays]
    meta_kinds = [_meta_kind.get(dt.kind, 'other') for dt in dtypes]
    # mixing string and numeric => object
    if any(mk != meta_kinds[0] for mk in meta_kinds[1:]):
        return object
    elif meta_kinds[0] == 'numeric':
        return np.find_common_type(dtypes, [])
    elif meta_kinds[0] == 'str':
        need_unicode = any(dt.kind == 'U' for dt in dtypes)
        # unicode are coded with 4 bytes
        max_size = max(dt.itemsize // 4 if dt.kind == 'U' else dt.itemsize
                       for dt in dtypes)
        return np.dtype(('U' if need_unicode else 'S', max_size))
    else:
        return object


class LHDFStore:
    """Context manager for pandas HDFStore"""
    def __init__(self, filepath_or_buffer, **kwargs):
        if isinstance(filepath_or_buffer, pd.HDFStore):
            if not filepath_or_buffer.is_open:
                raise IOError('The HDFStore must be open for reading.')
            self.store = filepath_or_buffer
            self.close_store = False
        else:
            self.store = pd.HDFStore(filepath_or_buffer, **kwargs)
            self.close_store = True

    def __enter__(self):
        return self.store

    def __exit__(self, type_, value, traceback):
        if self.close_store:
            self.store.close()


class SequenceZip:
    """
    Represents the "combination" of several sequences.

    This is very similar to python's builtin zip but only accepts sequences and acts as a Sequence (it can be
    indexed and has a len).

    Parameters
    ----------
    sequences : Iterable of Sequence
        Sequences to combine.

    Examples
    --------
    >>> z = SequenceZip([['a', 'b', 'c'], [1, 2, 3]])
    >>> for i in range(len(z)):
    ...     print(z[i])
    ('a', 1)
    ('b', 2)
    ('c', 3)
    >>> for v in z:
    ...     print(v)
    ('a', 1)
    ('b', 2)
    ('c', 3)
    >>> list(z[1:4])
    [('b', 2), ('c', 3)]
    """
    def __init__(self, sequences):
        self.sequences = sequences
        length = len(sequences[0])
        bad_length_seqs = [i for i, s in enumerate(sequences[1:], start=1) if len(s) != length]
        if bad_length_seqs:
            first_bad = bad_length_seqs[0]
            raise ValueError(f"sequence {first_bad} has a length of {len(sequences[first_bad])} which is different "
                             f"from the length of the first sequence ({length})")
        self._length = length

    def __len__(self):
        return self._length

    def __getitem__(self, key):
        if isinstance(key, (int, np.integer)):
            return tuple(seq[key] for seq in self.sequences)
        else:
            assert isinstance(key, slice), f"key ({key}) has invalid type ({type(key)})"
            return SequenceZip([seq[key] for seq in self.sequences])

    def __iter__(self):
        return zip(*self.sequences)

    def __repr__(self):
        return f'SequenceZip({self.sequences})'


class Repeater:
    """
    Returns a virtual sequence with value repeated n times.
    The sequence is never actually created in memory.

    Parameters
    ----------
    value : any
        Value to repeat.
    n : int
        Number of times to repeat value.

    Notes
    -----
    This is very similar to itertools.repeat except this version returns a Sequence instead of an iterator,
    meaning it has a length and can be indexed.

    Examples
    --------
    >>> r = Repeater('a', 3)
    >>> list(r)
    ['a', 'a', 'a']
    >>> r[0]
    'a'
    >>> r[2]
    'a'
    >>> r[3]
    Traceback (most recent call last):
    ...
    IndexError: index out of range
    >>> r[-1]
    'a'
    >>> r[-3]
    'a'
    >>> r[-4]
    Traceback (most recent call last):
    ...
    IndexError: index out of range
    >>> len(r)
    3
    >>> list(r[1:])
    ['a', 'a']
    >>> list(r[:2])
    ['a', 'a']
    >>> list(r[10:])
    []
    """
    def __init__(self, value, n):
        self.value = value
        self.n = n

    def __len__(self):
        return self.n

    def __getitem__(self, key):
        if isinstance(key, (int, np.integer)):
            if key >= self.n or key < -self.n:
                raise IndexError('index out of range')
            return self.value
        else:
            assert isinstance(key, slice), f"key ({key}) has invalid type ({type(key)})"
            start, stop, step = key.indices(self.n)
            # XXX: unsure // step is correct
            return Repeater(self.value, (stop - start) // step)

    def __iter__(self):
        return itertools.repeat(self.value, self.n)

    def __repr__(self):
        return f'Repeater({self.value}, {self.n})'


# TODO: remove Product from larray_editor.utils (it is almost identical)
class Product:
    """
    Represents the `cartesian product` of several sequences.

    This is very similar to itertools.product but only accepts sequences and acts as a sequence (it can be
    indexed and has a len).

    Parameters
    ----------
    sequences : Iterable of Sequence
        Sequences on which to apply the cartesian product.

    Examples
    --------
    >>> p = Product([['a', 'b', 'c'], [1, 2]])
    >>> for i in range(len(p)):
    ...     print(p[i])
    ('a', 1)
    ('a', 2)
    ('b', 1)
    ('b', 2)
    ('c', 1)
    ('c', 2)
    >>> p[1:4]
    [('a', 2), ('b', 1), ('b', 2)]
    >>> p[-3:]
    [('b', 2), ('c', 1), ('c', 2)]
    >>> list(p)
    [('a', 1), ('a', 2), ('b', 1), ('b', 2), ('c', 1), ('c', 2)]
    >>> list(Product([['a', 'b', 'c']]))
    [('a',), ('b',), ('c',)]
    >>> list(Product([]))
    [()]
    """
    def __init__(self, sequences):
        self.sequences = sequences
        shape = [len(a) for a in self.sequences]
        self._div_mod = [(int(np.prod(shape[i + 1:])), shape[i])
                         for i in range(len(shape))]
        self._length = np.prod(shape, dtype=int)

    def __len__(self):
        return self._length

    def __getitem__(self, key):
        if isinstance(key, (int, np.integer)):
            if key >= self._length:
                raise IndexError(f"index {key} out of range for Product of length {self._length}")
            # this is similar to np.unravel_index but a tad faster for scalars
            return tuple(array[key // div % mod]
                         for array, (div, mod) in zip(self.sequences, self._div_mod))
        else:
            assert isinstance(key, slice), f"key ({key}) has invalid type ({type(key)})"
            start, stop, step = key.indices(self._length)
            div_mod = self._div_mod
            arrays = self.sequences
            # XXX: we probably want to return another Product object with an updated start/stop to stay
            #      lazy in that case too.
            return [tuple(array[idx // div % mod]
                          for array, (div, mod) in zip(arrays, div_mod))
                    for idx in range(start, stop, step)]

    def __iter__(self):
        return product(*self.sequences)

    def __repr__(self):
        return f'Product({self.sequences})'


_np_generic = np.generic


def _kill_np_type(value):
    return value.item() if isinstance(value, _np_generic) else value


_kill_np_types = np.vectorize(_kill_np_type, otypes=[object])


def ensure_no_numpy_type(array):
    """
    Converts array to a (potentially nested) list of builtin Python values (i.e. using no numpy-specific types)

    Parameters
    ----------
    array : np.ndarray
        array to convert.

    Returns
    -------
    list
        a (potentially nested) list with the same "shape" as `array` but any numpy type converted to the closest
        Python builtin type
    """
    assert isinstance(array, np.ndarray)
    if array.dtype.kind == 'O':
        array = _kill_np_types(array)
    return array.tolist()


# ################# #
#  validator funcs  #
# ################# #

def _positive_integer(value):
    if not (isinstance(value, int) and value > 0):
        raise ValueError("Expected positive integer")


def _validate_dir(directory):
    if not os.path.isdir(directory):
        raise ValueError(f"The directory {directory} could not be found")


def argsort(seq):
    return sorted(range(len(seq)), key=seq.__getitem__)


def exactly_one(a: bool, b: bool, c: bool = False) -> bool:
    """returns True if exactly one of a, b or c boolean arguments is True, False otherwise"""
    return (a or b) and not (a and b) if not c else not (a or b)



import os
import atexit

import numpy as np
try:
    import xlwings as xw
except ImportError:
    xw = None

from larray.core.array import Array, ndtest             # noqa: F401
from larray.core.axis import Axis
from larray.core.constants import nan
from larray.core.group import _translate_sheet_name
from larray.inout.pandas import df_asarray
from larray.inout.misc import from_lists
from larray.util.misc import deprecate_kwarg


string_types = (str,)


if xw is not None:
    from xlwings.conversion.pandas_conv import PandasDataFrameConverter

    from xlwings.constants import FileFormat
    global_app = None

    def is_app_alive(app):
        try:
            app.books
            return True
        except Exception:
            return False

    def kill_global_app():
        global global_app

        if global_app is not None:
            if is_app_alive(global_app):
                try:
                    global_app.quit()
                except Exception:
                    pass
            del global_app
            global_app = None

    class ArrayConverter(PandasDataFrameConverter):
        writes_types = Array

        @classmethod
        def read_value(cls, value, options):
            df = PandasDataFrameConverter.read_value(value, options)
            return df_asarray(df)

        @classmethod
        def write_value(cls, value, options):
            df = value.to_frame(fold_last_axis_name=True)
            return PandasDataFrameConverter.write_value(df, options)

    ArrayConverter.register(Array)

    def _disable_screen_updates(app):
        xl_app = app.api
        xl_app.ScreenUpdating = False
        xl_app.DisplayStatusBar = False
        # this makes our test suite freeze
        # app.calculation = "manual"
        # unsure we can safely do this
        # xl_app.EnableEvents = False

    class Workbook:
        # TODO: replace overwrite_file by mode='r'|'w'|'a' the day xlwings will support a read-only mode
        def __init__(self, filepath=None, overwrite_file=False, visible=None, silent=None, app=None, load_addins=None):
            global global_app

            xw_wkb = None
            self.delayed_filepath = None
            self.filepath = None
            self.new_workbook = False
            self.active_workbook = filepath == -1

            if filepath is None:
                self.new_workbook = True

            if isinstance(filepath, str):
                basename, ext = os.path.splitext(filepath)
                if ext:
                    # XXX: we might want to be more precise than .xl* because I am unsure writing .xls
                    #     (or anything other than .xlsx and .xlsm) would work
                    if not ext.startswith('.xl'):
                        raise ValueError(f"'{ext}' is not a supported file extension")
                    if not os.path.isfile(filepath) and not overwrite_file:
                        raise ValueError(f"File {filepath} does not exist. Please give the path to an existing file "
                                         f"or set overwrite_file argument to True")
                    if os.path.isfile(filepath) and overwrite_file:
                        self.filepath = filepath
                        # we create a temporary file to work on. In case of crash, the original is not destroyed.
                        # the temporary file is renamed as the original file at close.
                        filepath = basename + '~' + ext
                    if not os.path.isfile(filepath):
                        self.new_workbook = True
                else:
                    # try to target an open but unsaved workbook. We cannot use the same code path as for other options
                    # because we do not know which Excel instance has that book
                    xw_wkb = xw.Book(filepath)
                    app = xw_wkb.app

            # active workbook use active app by default
            if self.active_workbook and app not in {None, "active"}:
                raise ValueError("to connect to the active workbook, one must use the 'active' Excel instance "
                                 "(app='active' or app=None)")

            # unless explicitly set, app is set to visible for brand new or active book.
            # For unsaved_book it is left intact.
            if visible is None:
                if filepath is None or self.active_workbook:
                    visible = True
                elif xw_wkb is None:
                    # filepath is not None and we target a real file (not an unsaved book)
                    visible = False

            if app is None:
                if self.active_workbook:
                    app = "active"
                elif visible:
                    app = "new"
                else:
                    app = "global"

            if load_addins is None:
                load_addins = visible and app == "new"

            if app == "new":
                app = xw.App(visible=visible, add_book=False)
                if not visible:
                    _disable_screen_updates(app)
            elif app == "active":
                app = xw.apps.active
            elif app == "global":
                if global_app is None:
                    atexit.register(kill_global_app)
                if global_app is None or not is_app_alive(global_app):
                    global_app = xw.App(visible=visible, add_book=False)
                    if not visible:
                        _disable_screen_updates(global_app)
                app = global_app
            assert isinstance(app, xw.App)

            # activate XLA(M) addins, if nee
            # By default, add-ins are not activated when an Excel Workbook is opened via COM
            if load_addins:
                xl_app = app.api
                for i in range(1, xl_app.AddIns.Count + 1):
                    addin = xl_app.AddIns(i)
                    addin_path = addin.FullName
                    if addin.Installed and '.xll' not in addin_path.lower():
                        xl_app.Workbooks.Open(addin_path)

            if visible:
                app.visible = visible

            if silent is None:
                silent = not visible

            update_links_backup = app.api.AskToUpdateLinks
            display_alerts_backup = app.display_alerts
            if silent:
                # try to update links silently instead of asking: "Update", "Don't Update", "Help"
                app.api.AskToUpdateLinks = False

                # in case some links cannot be updated, continue instead of asking: "Continue" or "Edit Links..."
                app.display_alerts = False

            if filepath is None:
                # creates a new/blank Book
                xw_wkb = app.books.add()
            elif self.active_workbook:
                xw_wkb = app.books.active
            elif xw_wkb is None:
                # file already exists (and is a file)
                if os.path.isfile(filepath):
                    xw_wkb = app.books.open(filepath)
                else:
                    # let us remember the path
                    self.delayed_filepath = filepath
                    xw_wkb = app.books.add()

            if silent:
                app.api.AskToUpdateLinks = update_links_backup
                app.display_alerts = display_alerts_backup

            self.xw_wkb = xw_wkb

        @property
        def app(self):
            return self.xw_wkb.app

        def __contains__(self, key):
            if isinstance(key, int):
                length = len(self)
                return -length <= key < length
            else:
                # I would like to use: "return key in wb.sheets" but as of xlwings 0.10 wb.sheets.__contains__ does not
                # work for sheet names (it works with Sheet objects I think)
                return key in self.sheet_names()

        def _ipython_key_completions_(self):
            return list(self.sheet_names())

        def __getitem__(self, key):
            key = _translate_sheet_name(key)
            if key in self:
                return Sheet(self, key)
            else:
                raise KeyError(f'Workbook has no sheet named {key}')

        def __setitem__(self, key, value):
            key = _translate_sheet_name(key)
            if self.new_workbook:
                self.xw_wkb.sheets[0].name = key
                self.new_workbook = False
            key_in_self = key in self
            if isinstance(value, Sheet):
                if value.xw_sheet.book.app != self.app:
                    raise ValueError("cannot copy a sheet from one instance of Excel to another")

                # xlwings index is 1-based
                # TODO: implement Workbook.index(key)
                target_idx = self[key].xw_sheet.index - 1 if key_in_self else -1
                target_sheet = self[target_idx].xw_sheet
                # add new sheet after target sheet. The new sheet will be named something like "value.name (1)" but I
                # do not think there is anything we can do about this, except rename it afterwards because Copy has no
                # name argument. See https://msdn.microsoft.com/en-us/library/office/ff837784.aspx
                value.xw_sheet.api.Copy(None, target_sheet.api)
                if key_in_self:
                    target_sheet.delete()
                # rename the new sheet
                self[target_idx].name = key
                return
            if key_in_self:
                sheet = self[key]
                sheet.clear()
            else:
                xw_sheet = self.xw_wkb.sheets.add(key, after=self[-1].xw_sheet)
                sheet = Sheet(None, None, xw_sheet=xw_sheet)
            sheet["A1"] = value

        def __delitem__(self, key):
            self[key].delete()

        def sheet_names(self):
            return [s.name for s in self]

        def save(self, path=None, password=None):
            r"""Save Workbook to file.

            Parameters
            ----------
            path : str, optional
                Path to save the file to. Defaults to None (use the path used when opening the workbook).
            password : str, optional
                Password to protect the file. Defaults to None (no password).
            """
            # saved_path = self.xw_wkb.api.Path
            # was_saved = saved_path != ''
            if path is None and self.delayed_filepath is not None:
                path = self.delayed_filepath

            if password is not None:
                if path is None:
                    raise ValueError("saving a Workbook with a password is only supported for workbooks with an "
                                     "explicit path (given either when opening the workbook or here as the path "
                                     "argument)")
                realpath = os.path.realpath(path)
                # XXX: this is probably Windows only
                # using Password as keyword argument does not work !
                self.xw_wkb.api.SaveAs(realpath, FileFormat.xlOpenXMLWorkbook, password)
            else:
                self.xw_wkb.save(path=path)

        def close(self):
            # Close the workbook in Excel.
            # This will not quit the Excel instance, even if this was the last workbook of that Excel instance.
            if self.filepath is not None and os.path.isfile(self.xw_wkb.fullname):
                tmp_file = self.xw_wkb.fullname
                self.xw_wkb.close()
                # XXX: do we check for this case earlier and act differently depending on overwrite?
                os.remove(self.filepath)
                os.rename(tmp_file, self.filepath)
            else:
                self.xw_wkb.close()

        def __iter__(self):
            return iter([Sheet(None, None, xw_sheet)
                         for xw_sheet in self.xw_wkb.sheets])

        def __len__(self):
            return len(self.xw_wkb.sheets)

        def __dir__(self):
            return list(set(dir(self.__class__)) | set(dir(self.xw_wkb)))

        def __getattr__(self, key):
            return getattr(self.xw_wkb, key)

        def __enter__(self):
            return self

        def __exit__(self, type_, value, traceback):
            # XXX: we should probably also avoid closing the workbook for visible=True???
            # XXX: we might want to disallow using open_excel as a context manager (in __enter__)
            #      when we have nothing to do in close because it is kinda misleading (this might piss off
            #      users though, so maybe a warning would be better).
            if not self.active_workbook:
                self.close()

        def __repr__(self):
            cls = self.__class__
            return f'<{cls.__module__}.{cls.__name__} [{self.name}]>'

    def _fill_slice(s, length):
        r"""
        replaces a slice None bounds by actual bounds.

        Parameters
        ----------
        s : slice
            slice to replace
        length : int
            length of sequence

        Returns
        -------
        slice
        """
        return slice(s.start if s.start is not None else 0, s.stop if s.stop is not None else length, s.step)

    def _concrete_key(key, obj, ndim=2):
        r"""Expand key to ndim and replace None in slices start/stop bounds by 0 or obj.shape[corresponding_dim]
        respectively.

        Parameters
        ----------
        key : scalar, slice or tuple
            input key
        obj : object
            any object with a 'shape' attribute.
        ndim : int
            number of dimensions to expand to. We could use len(obj.shape) instead but we avoid it to not trigger
            obj.shape, which can be expensive in the case of a sheet with blank cells after the data.
        """
        if not isinstance(key, tuple):
            key = (key,)

        if len(key) < ndim:
            key = key + (slice(None),) * (ndim - len(key))

        # only compute shape if necessary because it can be expensive in some cases
        if any(isinstance(k, slice) and k.stop is None for k in key):
            shape = obj.shape
        else:
            shape = (None, None)

        # We use _fill_slice instead of slice(*k.indices(length)) because the later also clips bounds which exceed
        # the length and we do NOT want to do that in this case (see issue #273).
        return [_fill_slice(k, length) if isinstance(k, slice) else k
                for k, length in zip(key, shape)]

    class Sheet:
        def __init__(self, workbook, key, xw_sheet=None):
            if xw_sheet is None:
                xw_sheet = workbook.xw_wkb.sheets[key]
            object.__setattr__(self, 'xw_sheet', xw_sheet)

        # TODO: we can probably scrap this for xlwings 0.9+. We need to have
        #       a unit test for this though.
        def __getitem__(self, key):
            if isinstance(key, string_types):
                return Range(self, key)

            row, col = _concrete_key(key, self)
            if isinstance(row, slice) or isinstance(col, slice):
                row1, row2 = (row.start, row.stop) if isinstance(row, slice) else (row, row + 1)
                col1, col2 = (col.start, col.stop) if isinstance(col, slice) else (col, col + 1)
                return Range(self, (row1 + 1, col1 + 1), (row2, col2))
            else:
                return Range(self, (row + 1, col + 1))

        def __setitem__(self, key, value):
            if isinstance(value, Array):
                value = value.dump(header=False)
            self[key].xw_range.value = value

        @property
        def shape(self):
            r"""
            shape of sheet including top-left empty rows/columns but excluding bottom-right ones.
            """
            from xlwings.constants import Direction as xldir

            sheet = self.xw_sheet.api
            used = sheet.UsedRange
            first_row = used.Row
            first_col = used.Column
            last_row = first_row + used.Rows.Count - 1
            last_col = first_col + used.Columns.Count - 1
            last_cell = sheet.Cells(last_row, last_col)

            # fast path for sheets with a non blank bottom-right value
            if last_cell.Value is not None:
                return last_row, last_col

            last_row_used = last_cell.End(xldir.xlToLeft).Value is not None
            last_col_used = last_cell.End(xldir.xlUp).Value is not None

            # fast path for sheets where last row and last col are not entirely blank
            if last_row_used and last_col_used:
                return last_row, last_col
            else:
                LEFT, UP = xldir.xlToLeft, xldir.xlUp

                def line_length(row, col, direction):
                    last_cell = sheet.Cells(row, col)
                    if last_cell.Value is not None:
                        return col if direction is LEFT else row
                    first_cell = last_cell.End(direction)
                    pos = first_cell.Column if direction is LEFT else first_cell.Row
                    return pos - 1 if first_cell.Value is None else pos

                if last_row < last_col:
                    if last_row_used or last_row == 1:
                        max_row = last_row
                    else:
                        for max_row in range(last_row - 1, first_row - 1, -1):
                            if line_length(max_row, last_col, LEFT) > 0:
                                break
                    if last_col_used or last_col == 1:
                        max_col = last_col
                    else:
                        max_col = max(line_length(row, last_col, LEFT) for row in range(first_row, max_row + 1))
                else:
                    if last_col_used or last_col == 1:
                        max_col = last_col
                    else:
                        for max_col in range(last_col - 1, first_col - 1, -1):
                            if line_length(last_row, max_col, UP) > 0:
                                break
                    if last_row_used or last_row == 1:
                        max_row = last_row
                    else:
                        max_row = max(line_length(last_row, col, UP) for col in range(first_col, max_col + 1))
                return max_row, max_col

        @property
        def ndim(self):
            return 2

        def __array__(self, dtype=None):
            return np.asarray(self[:], dtype=dtype)

        def __dir__(self):
            return list(set(dir(self.__class__)) | set(dir(self.xw_sheet)))

        def __getattr__(self, key):
            return getattr(self.xw_sheet, key)

        def __setattr__(self, key, value):
            setattr(self.xw_sheet, key, value)

        @deprecate_kwarg('nb_index', 'nb_axes', arg_converter=lambda x: x + 1)
        def load(self, header=True, convert_float=True, nb_axes=None, index_col=None, fill_value=nan,
                 sort_rows=False, sort_columns=False, wide=True):
            return self[:].load(header=header, convert_float=convert_float, nb_axes=nb_axes, index_col=index_col,
                                fill_value=fill_value, sort_rows=sort_rows, sort_columns=sort_columns, wide=wide)

        # TODO: generalize to more than 2 dimensions or scrap it
        def array(self, data, row_labels=None, column_labels=None, names=None):
            r"""

            Parameters
            ----------
            data : str
                range for data
            row_labels : str, optional
                range for row labels
            column_labels : str, optional
                range for column labels
            names : list of str, optional

            Returns
            -------
            Array
            """
            if row_labels is not None:
                row_labels = np.asarray(self[row_labels])
            if column_labels is not None:
                column_labels = np.asarray(self[column_labels])
            if names is not None:
                labels = (row_labels, column_labels)
                axes = [Axis(axis_labels, name) for axis_labels, name in zip(labels, names)]
            else:
                axes = (row_labels, column_labels)
            # _converted_value is used implicitly via Range.__array__
            return Array(np.asarray(self[data]), axes)

        def __repr__(self):
            cls = self.__class__
            xw_sheet = self.xw_sheet
            return f'<{cls.__module__}.{cls.__name__} [{xw_sheet.book.name}]{xw_sheet.name}>'

    class Range:
        def __init__(self, sheet, *args):
            xw_range = sheet.xw_sheet.range(*args)

            object.__setattr__(self, 'sheet', sheet)
            object.__setattr__(self, 'xw_range', xw_range)

        def _range_key_to_sheet_key(self, key):
            # string keys does not make sense in this case
            assert not isinstance(key, string_types)
            row_offset = self.xw_range.row1 - 1
            col_offset = self.xw_range.col1 - 1
            row, col = _concrete_key(key, self.xw_range)
            row = slice(row.start + row_offset, row.stop + row_offset) if isinstance(row, slice) else row + row_offset
            col = slice(col.start + col_offset, col.stop + col_offset) if isinstance(col, slice) else col + col_offset
            return row, col

        # TODO: we can probably scrap this for xlwings 0.9+. We need to have
        #       a unit test for this though.
        def __getitem__(self, key):
            return self.sheet[self._range_key_to_sheet_key(key)]

        def __setitem__(self, key, value):
            self.sheet[self._range_key_to_sheet_key(key)] = value

        def _converted_value(self, convert_float=True):
            list_data = self.xw_range.value

            # As of version 0.7.2 of xlwings, there is no built-in converter for
            # this. The builtin .options(numbers=int) converter converts all
            # values to int, whether that would loose information or not, but
            # this is not what we want.
            if convert_float:
                # Excel 'numbers' are always floats
                def convert(value):
                    if isinstance(value, float):
                        int_val = int(value)
                        if int_val == value:
                            return int_val
                        return value
                    elif isinstance(value, list):
                        return [convert(v) for v in value]
                    else:
                        return value
                return convert(list_data)
            return list_data

        def __float__(self):
            # no need to use _converted_value because we will convert back to a float anyway
            return float(self.xw_range.value)

        def __int__(self):
            # no need to use _converted_value because we will convert to an int anyway
            return int(self.xw_range.value)

        def __index__(self):
            v = self._converted_value()
            if hasattr(v, '__index__'):
                return v.__index__()
            else:
                raise TypeError("only integer scalars can be converted to a scalar index")

        def __array__(self, dtype=None):
            return np.array(self._converted_value(), dtype=dtype)

        def __larray__(self):
            return Array(self._converted_value())

        def __dir__(self):
            return list(set(dir(self.__class__)) | set(dir(self.xw_range)))

        def __getattr__(self, key):
            if hasattr(Array, key):
                return getattr(self.__larray__(), key)
            else:
                return getattr(self.xw_range, key)

        def __setattr__(self, key, value):
            setattr(self.xw_range, key, value)

        # TODO: implement all binops
        # def __mul__(self, other):
        #     return self.__larray__() * other

        def __str__(self):
            return str(self.__larray__())
        __repr__ = __str__

        @deprecate_kwarg('nb_index', 'nb_axes', arg_converter=lambda x: x + 1)
        def load(self, header=True, convert_float=True, nb_axes=None, index_col=None, fill_value=nan,
                 sort_rows=False, sort_columns=False, wide=True):
            if not self.ndim:
                return Array([])

            list_data = self._converted_value(convert_float=convert_float)

            if header:
                return from_lists(list_data, nb_axes=nb_axes, index_col=index_col, fill_value=fill_value,
                                  sort_rows=sort_rows, sort_columns=sort_columns, wide=wide)
            else:
                return Array(list_data)

    # XXX: deprecate this function?
    def open_excel(filepath=None, overwrite_file=False, visible=None, silent=None, app=None, load_addins=None):
        return Workbook(filepath, overwrite_file=overwrite_file, visible=visible, silent=silent, app=app,
                        load_addins=load_addins)
else:
    class Workbook:
        def __init__(self, filepath=None, overwrite_file=False, visible=None, silent=None, app=None, load_addins=None):
            raise Exception("Workbook class cannot be instantiated because xlwings is not installed")

        def app(self):
            raise Exception()

        def sheet_names(self):
            raise Exception()

        def save(self, path=None):
            raise Exception()

        def close(self):
            raise Exception()

    def open_excel(filepath=None, overwrite_file=False, visible=None, silent=None, app=None, load_addins=None):
        raise Exception("open_excel() is not available because xlwings is not installed")


# We define Workbook and open_excel documentation here since Readthedocs runs on Linux
Workbook.__doc__ = r"""
Excel Workbook.

See Also
--------
open_excel
"""

Workbook.sheet_names.__doc__ = r"""
Returns the names of the Excel sheets.

Examples
--------
>>> arr, arr2, arr3 = ndtest((3, 3)), ndtest((2, 2)), ndtest(4)
>>> with open_excel('excel_file.xlsx', overwrite_file=True) as wb:   # doctest: +SKIP
...     wb['arr'] = arr.dump()
...     wb['arr2'] = arr2.dump()
...     wb['arr3'] = arr3.dump()
...     wb.save()
...
...     wb.sheet_names()
['arr', 'arr2', 'arr3']
"""

Workbook.save.__doc__ = r"""
Saves the Workbook.

If a path is being provided, this works like SaveAs() in Excel.
If no path is specified and if the file hasn’t been saved previously,
it’s being saved in the current working directory with the current filename.
Existing files are overwritten without prompting.

Parameters
----------
path : str, optional
    Full path to the workbook. Defaults to None.

Examples
--------
>>> arr, arr2, arr3 = ndtest((3, 3)), ndtest((2, 2)), ndtest(4)
>>> with open_excel('excel_file.xlsx', overwrite_file=True) as wb:   # doctest: +SKIP
...     wb['arr'] = arr.dump()
...     wb['arr2'] = arr2.dump()
...     wb['arr3'] = arr3.dump()
...     wb.save()
"""

Workbook.close.__doc__ = r"""
Close the workbook in Excel.

Need to be called if the workbook has been opened without the `with` statement.

Examples
--------
>>> arr, arr2, arr3 = ndtest((3, 3)), ndtest((2, 2)), ndtest(4)   # doctest: +SKIP
>>> wb = open_excel('excel_file.xlsx', overwrite_file=True)       # doctest: +SKIP
>>> wb['arr'] = arr.dump()                                        # doctest: +SKIP
>>> wb['arr2'] = arr2.dump()                                      # doctest: +SKIP
>>> wb['arr3'] = arr3.dump()                                      # doctest: +SKIP
>>> wb.save()                                                     # doctest: +SKIP
>>> wb.close()                                                    # doctest: +SKIP
"""

Workbook.app.__doc__ = r"""
Return the Excel instance this workbook is attached to.
"""

open_excel.__doc__ = r"""
Open an Excel workbook

Parameters
----------
filepath : None, int or str, optional
    path to the Excel file. The file must exist if overwrite_file is False. Use None for a new blank workbook,
    -1 for the currently active workbook. Defaults to None.
overwrite_file : bool, optional
    whether or not to overwrite an existing file, if any. Defaults to False.
visible : None or bool, optional
    whether or not Excel should be visible. Defaults to False for files, True for new/active workbooks and to None
    ("unchanged") for existing unsaved workbooks.
silent : None or bool, optional
    whether or not to show dialog boxes for updating links or when some links cannot be updated.
    Defaults to False if visible, True otherwise.
app : None, "new", "active", "global" or xlwings.App, optional
    use "new" for opening a new Excel instance, "active" for the last active instance (including ones opened by the
    user) and "global" to (re)use the same instance for all workbooks of a program. None is equivalent to "active" if
    filepath is -1, "new" if visible is True and "global" otherwise. Defaults to None.

    The "global" instance is a specific Excel instance for all input from/output to Excel from within a single Python
    program (and should not interact with instances manually opened by the user or another program).
load_addins : None or bool, optional
    whether or not to load Excel addins. Defaults to True if visible and app == "new", False otherwise.

Returns
-------
Excel workbook.

Examples
--------
>>> arr = ndtest((3, 3))
>>> arr
a\b  b0  b1  b2
 a0   0   1   2
 a1   3   4   5
 a2   6   7   8

create a new Excel file and save an array

>>> # to create a new Excel file, argument overwrite_file must be set to True
>>> with open_excel('excel_file.xlsx', overwrite_file=True) as wb:   # doctest: +SKIP
...     wb['arr'] = arr.dump()
...     wb.save()

read array from an Excel file

>>> with open_excel('excel_file.xlsx') as wb:    # doctest: +SKIP
...     arr2 = wb['arr'].load()
>>> arr2    # doctest: +SKIP
a\b  b0  b1  b2
 a0   0   1   2
 a1   3   4   5
 a2   6   7   8
"""



import warnings

import numpy as np
import pandas as pd
from pandas import HDFStore

from larray.core.array import Array
from larray.core.axis import Axis
from larray.core.constants import nan
from larray.core.group import Group, LGroup, _translate_group_key_hdf
from larray.core.metadata import Metadata
from larray.util.misc import LHDFStore
from larray.inout.session import register_file_handler
from larray.inout.common import FileHandler, _supported_typenames, _supported_scalars_types
from larray.inout.pandas import df_asarray
from larray.example import get_example_filepath         # noqa: F401


# for backward compatibility (larray < 0.29) but any object read from an hdf file should have
# an attribute 'type'
def _get_type_from_attrs(attrs):
    return attrs.type if 'type' in attrs else 'Array'


def read_hdf(filepath_or_buffer, key, fill_value=nan, na=nan, sort_rows=False, sort_columns=False,
             name=None, **kwargs):
    r"""Reads a scalar or an axis or group or array named key from a HDF5 file in filepath (path+name)

    Parameters
    ----------
    filepath_or_buffer : str or pandas.HDFStore
        Path and name where the HDF5 file is stored or a HDFStore object.
    key : str or Group
        Name of the scalar or axis or group or array.
    fill_value : scalar or Array, optional
        Value used to fill cells corresponding to label combinations which are not present in the input.
        Defaults to NaN.
    sort_rows : bool, optional
        Whether or not to sort the rows alphabetically.
        Must be False if the read array has been dumped with an larray version >= 0.30.
        Defaults to False.
    sort_columns : bool, optional
        Whether or not to sort the columns alphabetically.
        Must be False if the read array has been dumped with an larray version >= 0.30.
        Defaults to False.
    name : str, optional
        Name of the axis or group to return. If None, name is set to passed key.
        Defaults to None.

    Returns
    -------
    Array

    Examples
    --------
    >>> fname = get_example_filepath('examples.h5')

    Read array by passing its identifier (key) inside the HDF file

    >>> # The data below is derived from a subset of the demo_pjan table from Eurostat
    >>> read_hdf(fname, 'pop')                     # doctest: +SKIP
    country  gender\time      2013      2014      2015
    Belgium         Male   5472856   5493792   5524068
    Belgium       Female   5665118   5687048   5713206
     France         Male  31772665  32045129  32174258
     France       Female  33827685  34120851  34283895
    Germany         Male  39380976  39556923  39835457
    Germany       Female  41142770  41210540  41362080
    """
    if not np.isnan(na):
        fill_value = na
        warnings.warn("read_hdf `na` argument has been renamed to `fill_value`. Please use that instead.",
                      FutureWarning, stacklevel=2)

    key = _translate_group_key_hdf(key)
    res = None
    with LHDFStore(filepath_or_buffer) as store:
        try:
            pd_obj = store.get(key)
        except KeyError:
            filepath = filepath_or_buffer if isinstance(filepath_or_buffer, HDFStore) else store.filename
            raise KeyError(f'No item with name {key} has been found in file {filepath}')
        attrs = store.get_storer(key).attrs
        writer = attrs.writer if 'writer' in attrs else None
        _type = _get_type_from_attrs(attrs)
        _meta = attrs.metadata if 'metadata' in attrs else None
        if _type == 'Array':
            # cartesian product is not necessary if the array was written by LArray
            cartesian_prod = writer != 'LArray'
            res = df_asarray(pd_obj, sort_rows=sort_rows, sort_columns=sort_columns, fill_value=fill_value,
                             parse_header=False, cartesian_prod=cartesian_prod)
            if _meta is not None:
                res.meta = _meta
        elif _type == 'Axis':
            if name is None:
                name = str(pd_obj.name)
            if name == 'None':
                name = None
            labels = pd_obj.values
            if 'dtype_kind' in attrs and attrs['dtype_kind'] == 'U':
                # this check is there because there are cases where dtype_kind is 'U' but pandas returns
                # an array with object dtype containing bytes instead of a string array, and in that case
                # np.char.decode does not work
                # this is at least the case for Python2 + Pandas 0.24.2 combination
                if labels.dtype.kind == 'O':
                    labels = np.array([label.decode('utf-8') for label in labels], dtype='U')
                else:
                    labels = np.char.decode(labels, 'utf-8')
            res = Axis(labels=labels, name=name)
            res._iswildcard = attrs['wildcard']
        elif _type == 'Group':
            if name is None:
                name = str(pd_obj.name)
            if name == 'None':
                name = None
            key = pd_obj.values
            if 'dtype_kind' in attrs and attrs['dtype_kind'] == 'U':
                key = np.char.decode(key, 'utf-8')
            axis = read_hdf(filepath_or_buffer, attrs['axis_key'])
            res = LGroup(key=key, name=name, axis=axis)
        elif _type in _supported_typenames:
            res = pd_obj.values
            assert len(res) == 1
            res = res[0]
    return res


@register_file_handler('pandas_hdf', ['h5', 'hdf'])
class PandasHDFHandler(FileHandler):
    r"""
    Handler for HDF5 files using Pandas.
    """
    def _open_for_read(self):
        self.handle = HDFStore(self.fname, mode='r')

    def _open_for_write(self):
        self.handle = HDFStore(self.fname)

    def list_items(self):
        keys = [key.strip('/') for key in self.handle.keys()]
        items = [(key, _get_type_from_attrs(self.handle.get_storer(key).attrs)) for key in keys if '/' not in key]
        # ---- for backward compatibility (LArray < 0.33) ----
        # axes
        items += [(key.split('/')[-1], 'Axis_Backward_Comp') for key in keys if '__axes__' in key]
        # groups
        items += [(key.split('/')[-1], 'Group_Backward_Comp') for key in keys if '__groups__' in key]
        return items

    def _read_item(self, key, typename, *args, **kwargs):
        if typename in _supported_typenames:
            hdf_key = '/' + key
        # ---- for backward compatibility (LArray < 0.33) ----
        elif typename == 'Axis_Backward_Comp':
            hdf_key = '__axes__/' + key
        elif typename == 'Group_Backward_Comp':
            hdf_key = '__groups__/' + key
        else:
            raise TypeError()
        return read_hdf(self.handle, hdf_key, *args, **kwargs)

    def _dump_item(self, key, value, *args, **kwargs):
        hdf_key = '/' + key
        if isinstance(value, (Array, Axis)):
            value.to_hdf(self.handle, hdf_key, *args, **kwargs)
        elif isinstance(value, Group):
            hdf_axis_key = '/' + value.axis.name
            value.to_hdf(self.handle, hdf_key, hdf_axis_key, *args, **kwargs)
        elif isinstance(value, _supported_scalars_types):
            s = pd.Series(data=value)
            self.handle.put(hdf_key, s)
            self.handle.get_storer(hdf_key).attrs.type = type(value).__name__
        else:
            raise TypeError()

    def _read_metadata(self):
        metadata = Metadata.from_hdf(self.handle)
        if metadata is None:
            metadata = Metadata()
        return metadata

    def _dump_metadata(self, metadata):
        metadata.to_hdf(self.handle)

    def close(self):
        self.handle.close()
